#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
#  All tools in this module have to be independent from DXF entities!
import enum
from typing import (
    List,
    Iterable,
    Tuple,
    TYPE_CHECKING,
    Union,
    Optional,
    Callable,
    NamedTuple,
)
import re
import math

from ezdxf.lldxf import validator, const
from ezdxf.lldxf.const import (
    MTextParagraphAlignment,
    MTextLineAlignment,
    MTextStroke,
    LEFT,
    CENTER,
    RIGHT,
    BASELINE,
    BOTTOM,
    MIDDLE,
    TOP,
)
from ezdxf.math import Vec3, Vec2, Vertex
from ezdxf.colors import rgb2int, RGB, int2rgb
from .fonts import FontMeasurements, AbstractFont, FontFace

if TYPE_CHECKING:
    from ezdxf.eztypes import Text, MText, DXFEntity, Tags

X_MIDDLE = 4  # special case for overall alignment "MIDDLE"

MTEXT_ALIGN_FLAGS = {
    1: (LEFT, TOP),
    2: (CENTER, TOP),
    3: (RIGHT, TOP),
    4: (LEFT, MIDDLE),
    5: (CENTER, MIDDLE),
    6: (RIGHT, MIDDLE),
    7: (LEFT, BOTTOM),
    8: (CENTER, BOTTOM),
    9: (RIGHT, BOTTOM),
}


class TextLine:
    """Helper class which represents a single line text entity
    (e.g. :class:`~ezdxf.entities.Text`).

    Args:
        text: content string
        font: ezdxf font definition like :class:`~ezdxf.tools.fonts.MonospaceFont`
            or :class:`~ezdxf.tools.fonts.MatplotlibFont`

    """

    def __init__(self, text: str, font: "AbstractFont"):
        self._font = font
        self._text_width: float = font.text_width(text)
        self._stretch_x: float = 1.0
        self._stretch_y: float = 1.0

    def stretch(self, alignment: str, p1: Vec3, p2: Vec3) -> None:
        """Set stretch factors for "FIT" and "ALIGNED" alignments to fit the
        text between `p1` and `p2`, only the distance between these points is
        important. Other given `alignment` values are ignore.

        """
        sx: float = 1.0
        sy: float = 1.0
        if alignment in ("FIT", "ALIGNED"):
            defined_length: float = (p2 - p1).magnitude
            if self._text_width > 1e-9:
                sx = defined_length / self._text_width
                if alignment == "ALIGNED":
                    sy = sx
        self._stretch_x = sx
        self._stretch_y = sy

    @property
    def width(self) -> float:
        """Returns the final (stretched) text width."""
        return self._text_width * self._stretch_x

    @property
    def height(self) -> float:
        """Returns the final (stretched) text height."""
        return self._font.measurements.total_height * self._stretch_y

    def font_measurements(self) -> FontMeasurements:
        """Returns the scaled font measurements."""
        return self._font.measurements.scale(self._stretch_y)

    def baseline_vertices(
        self,
        insert: Vertex,
        halign: int = 0,
        valign: int = 0,
        angle: float = 0,
        scale: Tuple[float, float] = (1, 1),
    ) -> List[Vec3]:
        """Returns the left and the right baseline vertex of the text line.

        Args:
            insert: insertion location
            halign: horizontal alignment left=0, center=1, right=2
            valign: vertical alignment baseline=0, bottom=1, middle=2, top=3
            angle: text rotation in radians
            scale: scale in x- and y-axis as 2-tuple of float

        """
        fm = self.font_measurements()
        vertices = [
            Vec2(0, fm.baseline),
            Vec2(self.width, fm.baseline),
        ]
        shift = self._shift_vector(halign, valign, fm)
        # Oblique angle is deliberately not supported, the base line should be
        # (near) the y-coordinate=0.
        return TextLine.transform_2d(vertices, insert, shift, angle, scale)

    def corner_vertices(
        self,
        insert: Vertex,
        halign: int = 0,
        valign: int = 0,
        angle: float = 0,
        scale: Tuple[float, float] = (1, 1),
        oblique: float = 0,
    ) -> List[Vec3]:
        """Returns the corner vertices of the text line in the order
        bottom left, bottom right, top right, top left.

        Args:
            insert: insertion location
            halign: horizontal alignment left=0, center=1, right=2
            valign: vertical alignment baseline=0, bottom=1, middle=2, top=3
            angle: text rotation in radians
            scale: scale in x- and y-axis as 2-tuple of float
            oblique: shear angle (slanting) in x-direction in radians

        """
        fm = self.font_measurements()
        vertices = [
            Vec2(0, fm.bottom),
            Vec2(self.width, fm.bottom),
            Vec2(self.width, fm.cap_top),
            Vec2(0, fm.cap_top),
        ]
        shift = self._shift_vector(halign, valign, fm)
        return TextLine.transform_2d(
            vertices, insert, shift, angle, scale, oblique
        )

    def _shift_vector(
        self, halign: int, valign: int, fm: FontMeasurements
    ) -> Tuple[float, float]:
        return _shift_x(self.width, halign), _shift_y(fm, valign)

    @staticmethod
    def transform_2d(
        vertices: Iterable[Vertex],
        insert: Vertex = Vec3(0, 0, 0),
        shift: Tuple[float, float] = (0, 0),
        rotation: float = 0,
        scale: Tuple[float, float] = (1, 1),
        oblique: float = 0,
    ) -> List[Vec3]:
        """Transform any vertices from the text line located at the base
        location at (0, 0) and alignment "LEFT".

        Args:
            vertices: iterable of vertices
            insert: insertion location
            shift: (shift-x, shift-y) as 2-tuple of float
            rotation: text rotation in radians
            scale: (scale-x, scale-y)  as 2-tuple of float
            oblique: shear angle (slanting) in x-direction in radians

        """
        # Building a transformation matrix vs. applying transformations in
        # individual steps:
        # Most text is horizontal, because people like to read horizontal text!
        # Operating in 2D is faster than building a full 3D transformation
        # matrix and a pure 2D transformation matrix is not implemented!
        # This function doesn't transform many vertices at the same time,
        # mostly only 4 vertices, therefore the matrix multiplication overhead
        # does not pay off.
        # The most expensive rotation transformation is the least frequently
        # used transformation.
        # IMPORTANT: this assumptions are not verified by profiling!

        # Use 2D vectors:
        vertices_: Iterable[Vec2] = Vec2.generate(vertices)

        # 1. slanting at the original location (very rare):
        if oblique:
            slant_x = math.tan(oblique)
            vertices_ = (Vec2(v.x + v.y * slant_x, v.y) for v in vertices_)

        # 2. apply alignment shifting (frequently):
        shift_vector = Vec2(shift)
        if shift_vector:
            vertices_ = (v + shift_vector for v in vertices_)

        # 3. scale (and mirror) at the aligned location (more often):
        scale_x, scale_y = scale
        if scale_x != 1 or scale_y != 1:
            vertices_ = (Vec2(v.x * scale_x, v.y * scale_y) for v in vertices_)

        # 4. apply rotation (rare):
        if rotation:
            vertices_ = (v.rotate(rotation) for v in vertices_)

        # 5. move to insert location in OCS/3D! (every time)
        insert = Vec3(insert)
        return [insert + v for v in vertices_]


def _shift_x(total_width: float, halign: int) -> float:
    if halign == CENTER:
        return -total_width / 2.0
    elif halign == RIGHT:
        return -total_width
    return 0.0  # LEFT


def _shift_y(fm: FontMeasurements, valign: int) -> float:
    if valign == BASELINE:
        return fm.baseline
    elif valign == MIDDLE:
        return -fm.cap_top + fm.cap_height / 2
    elif valign == X_MIDDLE:
        return -fm.cap_top + fm.total_height / 2
    elif valign == TOP:
        return -fm.cap_top
    return -fm.bottom


def unified_alignment(entity: Union["Text", "MText"]) -> Tuple[int, int]:
    """Return unified horizontal and vertical alignment.

    horizontal alignment: left=0, center=1, right=2

    vertical alignment: baseline=0, bottom=1, middle=2, top=3

    Returns:
        tuple(halign, valign)

    """
    dxftype = entity.dxftype()
    if dxftype in ("TEXT", "ATTRIB", "ATTDEF"):
        halign = entity.dxf.halign
        valign = entity.dxf.valign
        if halign in (3, 5):  # ALIGNED=3, FIT=5
            # For the alignments ALIGNED and FIT the text stretching has to be
            # handles separately.
            halign = CENTER
            valign = BASELINE
        elif halign == 4:  # MIDDLE is different to MIDDLE/CENTER
            halign = CENTER
            valign = X_MIDDLE
        return halign, valign
    elif dxftype == "MTEXT":
        return MTEXT_ALIGN_FLAGS.get(entity.dxf.attachment_point, (LEFT, TOP))
    else:
        raise TypeError(f"invalid DXF {dxftype}")


def plain_text(text: str) -> str:
    """Returns the plain text for :class:`~ezdxf.entities.Text`,
    :class:`~ezdxf.entities.Attrib` and :class:`~ezdxf.entities.Attdef` content.
    """
    # TEXT, ATTRIB and ATTDEF are short strings <= 255 in R12.
    # R2000 allows 2049 chars, but this limit is not often used in real world
    # applications.
    result = ""
    scanner = TextScanner(validator.fix_one_line_text(caret_decode(text)))
    while scanner.has_data:
        char = scanner.peek()
        if char == "%":  # special characters
            if scanner.peek(1) == "%":
                code = scanner.peek(2).lower()
                letter = const.SPECIAL_CHAR_ENCODING.get(code)
                if letter:
                    scanner.consume(3)  # %%?
                    result += letter
                    continue
                elif code in "kou":
                    # formatting codes (%%k, %%o, %%u) will be ignored in
                    # TEXT, ATTRIB and ATTDEF:
                    scanner.consume(3)
                    continue
        scanner.consume(1)
        # slightly faster then "".join(chars)
        result += char
    return result


ONE_CHAR_COMMANDS = "PNLlOoKkX"


##################################################
# MTEXT inline codes
# \L	Start underline
# \l	Stop underline
# \O	Start overline
# \o	Stop overline
# \K	Start strike-through
# \k	Stop strike-through
# \P	New paragraph (new line)
# \N	New column
# \~    None breaking space
# ^I    Tabulator
# \	Escape character - e.g. \\ = "\", \{ = "{"
#
# \p    start paragraph properties until next ";"
# \pi#,l#,r#; paragraph indent
#   i#  indent first line left, relative to (l)!
#   l#  indent paragraph left
#   r#  indent paragraph right
#   q?  alignments:
#   ql  align text in paragraph: left
#   qr  align text in paragraph: right
#   qc  align text in paragraph: center
#   qj  align text in paragraph: justified
#   qd  align text in paragraph: distributed
#   x   unknown meaning
#   t#[,c#,r#...] define absolute tabulator stops 1,c2,r3...
#       without prefix is a left adjusted tab stop
#       prefix 'c' for center adjusted tab stop
#       prefix 'r' for right adjusted tab stop
#   ?*  reset command to default value
#
# Examples:
# \pi1,t[5,20,...]; define tab stops as comma separated list
# \pxt4,c8,r12,16,c20,r24; left, centered and right adjusted tab stops
# \pi*,l*,r*,q*,t; reset to default values
# \pi2,l0;  = first line  2 & paragraph left 0
# \pi-2,l2; = first line -2 & paragraph left 2
# \pi0,l2;  = first line  0 & paragraph left 2
#
# \X	Paragraph wrap on the dimension line (only in dimensions)
# \Q	Slanting (oblique) text by angle - e.g. \Q30;
# \H	Text height relative - e.g. \H3x;
# \H	Text height absolute - e.g. \H3;
# \W	Text width factor - e.g. \W0.8x;
# \W	Text width absolute - e.g. \W0.8;
# \F	Font selection
# \f	Font selection
#
#     e.g. \Fgdt;o - GDT-tolerance
#     e.g. \fArial|b0|i0|c238|p10; - font Arial, non-bold, non-italic,
#     codepage 238, pitch 10
#     codepage 0 = no change
#     pitch 0 = no change
#
# \S	Stacking, fractions
#
#     e.g. \SA^ B;
#     A
#     B
#     e.g. \SX/ Y;
#     X
#     -
#     Y
#     e.g. \S1# 4;
#     1/4
#
# \A	Alignment relative to current line
#
#     \A0; = bottom
#     \A1; = center
#     \A2; = top
#
# \C	Color change
#
#     \C1; = red
#     \C2; = yellow
#     \C3; = green
#     \C4; = cyan
#     \C5; = blue
#     \C6; = magenta
#     \C7; = white

#     \c255; = RED (255, 0, 0)
#     \c65280; = GREEN (0, 255, 0)
#     \c16711680; = BLUE (0, 0, 255)

#     RGB color = \c7528479;  = 31,224,114
#     ezdxf.rgb2int((31,224,114)) = 2089074 (r,g,b) wrong!
#     ezdxf.rgb2int((114,224,31)) = 7528479 (b,g,r) reversed order is correct!
#
# \T	Tracking, char.spacing absolute - e.g. \T2;
# \T	Tracking, char.spacing relative - e.g. \T2x;
# {}	Braces - define the text area influenced by the code
#       Multiple codes after the opening brace are valid until the closing
#       brace.  e.g. {\H0.4x;\A1;small centered text}
#
# Codes and braces can be nested up to 8 levels deep
#
# Column types in BricsCAD:
#   - dynamic auto height: all columns have the same height
#   - dynamic manual height: each columns has an individual height
#   - no columns
#   - static: all columns have the same height, like dynamic auto height,
#     difference is only important for user interaction in CAD applications
#
# - All columns have the same width and gutter.
# - Paragraphs do overflow into the next column if required.


def fast_plain_mtext(text: str, split=False) -> Union[List[str], str]:
    """Returns the plain MTEXT content as a single string or  a list of
    strings if `split` is ``True``. Replaces ``\\P`` by ``\\n`` and removes
    other controls chars and inline codes.

    This function is more than 4x faster than :func:`plain_mtext`, but does not
    remove single letter inline commands with arguments without a terminating
    semicolon like this ``"\\C1red text"``.

    .. note::

        Well behaved CAD applications and libraries always create inline codes
        for commands with arguments with a terminating semicolon like this
        ``"\\C1;red text"``!

    Args:
        text: MTEXT content string
        split: split content at line endings ``\\P``

    """
    chars = []
    # split text into chars, in reversed order for efficient pop()
    raw_chars = list(reversed(text))
    while len(raw_chars):
        char = raw_chars.pop()
        if char == "\\":  # is a formatting command
            try:
                char = raw_chars.pop()
            except IndexError:
                break  # premature end of text - just ignore

            if char in "\\{}":
                chars.append(char)
            elif char in ONE_CHAR_COMMANDS:
                if char == "P":  # new line
                    chars.append("\n")
                elif char == "N":  # new column
                    # until columns are supported, better to at least remove the
                    # escape character
                    chars.append(" ")
                # else: discard other commands
            else:  # multiple character commands are terminated by ';'
                stacking = char == "S"  # stacking command surrounds user data
                first_char = char
                search_chars = raw_chars.copy()
                try:
                    while char != ";":  # end of format marker
                        char = search_chars.pop()
                        if stacking and char != ";":
                            # append user data of stacking command
                            chars.append(char)
                    raw_chars = search_chars
                except IndexError:
                    # premature end of text - just ignore
                    chars.append("\\")
                    chars.append(first_char)
        elif char in "{}":  # grouping
            pass  # discard group markers
        elif char == "%":  # special characters
            if len(raw_chars) and raw_chars[-1] == "%":
                raw_chars.pop()  # discard next '%'
                if len(raw_chars):
                    code = raw_chars.pop()
                    letter = const.SPECIAL_CHAR_ENCODING.get(code.lower())
                    if letter:
                        chars.append(letter)
                    else:
                        chars.extend(("%", "%", code))
            else:  # char is just a single '%'
                chars.append(char)
        else:  # char is what it is, a character
            chars.append(char)

    result = "".join(chars)
    return result.split("\n") if split else result


def caret_decode(text: str) -> str:
    """DXF stores some special characters using caret notation. This function
    decodes this notation to normalise the representation of special characters
    in the string.

    see: https://en.wikipedia.org/wiki/Caret_notation

    """

    def replace_match(match: "re.Match") -> str:
        c = ord(match.group(1))
        return chr((c - 64) % 126)

    return re.sub(r"\^(.)", replace_match, text)


def split_mtext_string(s: str, size: int = 250) -> List[str]:
    """Split the MTEXT content string into chunks of max `size`."""
    chunks = []
    pos = 0
    while True:
        chunk = s[pos : pos + size]
        if len(chunk):
            if len(chunk) < size:
                chunks.append(chunk)
                return chunks
            pos += size
            # do not split chunks at '^'
            if chunk[-1] == "^":
                chunk = chunk[:-1]
                pos -= 1
            chunks.append(chunk)
        else:
            return chunks


def plain_mtext(
    text: str,
    split=False,
    tabsize: int = 4,
) -> Union[List[str], str]:
    """Returns the plain MTEXT content as a single string or a list of
    strings if `split` is ``True``. Replaces ``\\P`` by ``\\n`` and removes
    other controls chars and inline codes.

    This function is much slower than :func:`fast_plain_mtext`, but removes all
    inline codes.

    Args:
        text: MTEXT content string
        split: split content at line endings ``\\P``
        tabsize: count of replacement spaces for tabulators ``^I``

    """
    content: List[str] = []
    paragraph: List[str] = []

    # localize enum to speed up inner loop
    (
        _,
        word,
        stack,
        space,
        nbsp,
        tabulator,
        new_paragraph,
        new_column,
        *_,
    ) = iter(TokenType)
    tab_replacement = " " * tabsize

    for token in MTextParser(text):
        t = token.type
        if t == word:
            paragraph.append(token.data)
        elif t == space or t == nbsp:
            paragraph.append(" ")
        elif t == new_paragraph or t == new_column:
            content.append("".join(paragraph))
            paragraph.clear()
        elif t == tabulator:
            paragraph.append(tab_replacement)
        elif t == stack:
            upr, lwr, divider = token.data
            paragraph.append(upr + divider + lwr)

    if paragraph:
        content.append("".join(paragraph))
    if split:
        return content
    else:
        return "\n".join(content)


def escape_dxf_line_endings(text: str) -> str:
    # replacing '\r\n' and '\n' by '\P' is required when exporting, else an
    # invalid DXF file would be created.
    return text.replace("\r", "").replace("\n", "\\P")


def replace_non_printable_characters(text: str, replacement: str = "▯") -> str:
    return "".join(replacement if is_non_printable_char(c) else c for c in text)


def is_non_printable_char(char: str) -> bool:
    return 0 <= ord(char) < 32 and char != "\t"


def text_wrap(
    text: str,
    box_width: Optional[float],
    get_text_width: Callable[[str], float],
) -> List[str]:
    """Wrap text at ``\\n`` and given `box_width`. This tool was developed for
    usage with the MTEXT entity. This isn't the most straightforward word
    wrapping algorithm, but it aims to match the behavior of AutoCAD.

    Args:
        text: text to wrap, included ``\\n`` are handled as manual line breaks
        box_width: wrapping length, ``None`` to just wrap at ``\\n``
        get_text_width: callable which returns the width of the given string

    """
    # Copyright (c) 2020-2021, Matthew Broadway
    # License: MIT License
    if not text or text.isspace():
        return []
    manual_lines = re.split(r"(\n)", text)  # includes \n as it's own token
    tokens = [t for line in manual_lines for t in re.split(r"(\s+)", line) if t]
    lines: List[str] = []
    current_line: str = ""
    line_just_wrapped = False

    for t in tokens:
        on_first_line = not lines
        if t == "\n" and line_just_wrapped:
            continue
        line_just_wrapped = False
        if t == "\n":
            lines.append(current_line.rstrip())
            current_line = ""
        elif t.isspace():
            if current_line or on_first_line:
                current_line += t
        else:
            if (
                box_width is not None
                and get_text_width(current_line + t) > box_width
            ):
                if not current_line:
                    current_line += t
                else:
                    lines.append(current_line.rstrip())
                    current_line = t
                    line_just_wrapped = True
            else:
                current_line += t

    if current_line and not current_line.isspace():
        lines.append(current_line.rstrip())
    return lines


def is_text_vertical_stacked(text: "DXFEntity") -> bool:
    """Returns ``True`` if the associated text :class:`~ezdxf.entities.Textstyle`
    is vertical stacked.
    """
    if not text.is_supported_dxf_attrib("style"):
        raise TypeError(
            f"{text.dxftype()} does not support the style attribute."
        )

    if text.doc:
        style = text.doc.styles.get(text.dxf.style)
        if style:
            return style.is_vertical_stacked  # type: ignore
    return False


_alignment_char = {
    MTextParagraphAlignment.DEFAULT: "",
    MTextParagraphAlignment.LEFT: "l",
    MTextParagraphAlignment.RIGHT: "r",
    MTextParagraphAlignment.CENTER: "c",
    MTextParagraphAlignment.JUSTIFIED: "j",
    MTextParagraphAlignment.DISTRIBUTED: "d",
}

COMMA = ","
DIGITS = "01234567890"


def rstrip0(s):
    if isinstance(s, (int, float)):
        return f"{s:g}"
    else:
        return s


class ParagraphProperties(NamedTuple):
    """Stores all known MTEXT paragraph properties in a :class:`NamedTuple`.
    Indentations and tab stops are multiples of the default text height
    :attr:`MText.dxf.char_height`. E.g. :attr:`char_height` is 0.25 and
    :attr:`indent` is 4, the real indentation is 4 x 0.25 = 1 drawing unit.
    The default tabulator stops are 4, 8, 12, ... if no tabulator stops are
    explicit defined.

    Args:
         indent (float): left indentation of the first line, relative to :attr:`left`,
            which means an :attr:`indent` of 0 has always the same indentation
            as :attr:`left`
         left (float): left indentation of the paragraph except for the first line
         right (float): left indentation of the paragraph
         align: :class:`~ezdxf.lldxf.const.MTextParagraphAlignment` enum
         tab_stops: tuple of tabulator stops, as ``float`` or as ``str``,
            ``float`` values are left aligned tab stops, strings with prefix
            ``"c"`` are center aligned tab stops and strings with prefix ``"r"``
            are right aligned tab stops

    """

    # Reset: \pi*,l*,r*,q*,t;
    indent: float = 0  # relative to left!
    left: float = 0
    right: float = 0
    align: MTextParagraphAlignment = MTextParagraphAlignment.DEFAULT
    # tab stops without prefix or numbers are left adjusted
    # tab stops, e.g 2 or '2'
    # prefix 'c' defines a center adjusted tab stop e.g. 'c3.5'
    # prefix 'r' defines a right adjusted tab stop e.g. 'r2.7'
    # The tab stop in drawing units = n x char_height
    tab_stops: Tuple = tuple()

    def tostring(self) -> str:
        """Returns the MTEXT paragraph properties as MTEXT inline code
        e.g. ``"\\pxi-2,l2;"``.

        """
        args = []
        if self.indent:
            args.append(f"i{self.indent:g}")
            args.append(COMMA)
        if self.left:
            args.append(f"l{self.left:g}")
            args.append(COMMA)
        if self.right:
            args.append(f"r{self.right:g}")
            args.append(COMMA)
        if self.align:
            args.append(f"q{_alignment_char[self.align]}")
            args.append(COMMA)
        if self.tab_stops:
            args.append(f"t{COMMA.join(map(rstrip0, self.tab_stops))}")
            args.append(COMMA)

        if args:
            if args[-1] == COMMA:
                args.pop()
            # exporting always "x" as second letter seems to be safe
            return "\\px" + "".join(args) + ";"
        else:
            return ""


# IMPORTANT for parsing MTEXT inline codes: "\\H0.1\\A1\\C1rot"
# Inline commands with a single argument, don't need a trailing ";"!


class MTextEditor:
    """The :class:`MTextEditor` is a helper class to build MTEXT content
    strings with support for inline codes to change color, font or
    paragraph properties. The result is always accessible by the :attr:`text`
    attribute or the magic :func:`__str__` function as
    :code:`str(MTextEditor("text"))`.

    All text building methods return `self` to implement a floating interface::

        e = MTextEditor("This example ").color("red").append("switches color to red.")
        mtext = msp.add_mtext(str(e))

    The initial text height, color, text style and so on is determined by the
    DXF attributes of the :class:`~ezdxf.entities.MText` entity.

    .. warning::

        The :class:`MTextEditor` assembles just the inline code, which has to be
        parsed and rendered by the target CAD application, `ezdxf` has no influence
        to that result.

        Keep inline formatting as simple as possible, don't test the limits of its
        capabilities, this will not work across different CAD applications and keep
        the formatting in a logic manner like, do not change paragraph properties
        in the middle of a paragraph.

        **There is no official documentation for the inline codes!**

    Args:
        text: init value of the MTEXT content string.

    """

    def __init__(self, text: str = ""):
        self.text = str(text)

    NEW_LINE = r"\P"
    NEW_PARAGRAPH = r"\P"
    NEW_COLUMN = r"\N"
    UNDERLINE_START = r"\L"
    UNDERLINE_STOP = r"\l"
    OVERSTRIKE_START = r"\O"
    OVERSTRIKE_STOP = r"\o"
    STRIKE_START = r"\K"
    STRIKE_STOP = r"\k"
    GROUP_START = "{"
    GROUP_END = "}"
    ALIGN_BOTTOM = r"\A0;"
    ALIGN_MIDDLE = r"\A1;"
    ALIGN_TOP = r"\A2;"
    NBSP = r"\~"  # non breaking space
    TAB = "^I"

    def append(self, text: str) -> "MTextEditor":
        """Append `text`."""
        self.text += text
        return self

    def __iadd__(self, text: str) -> "MTextEditor":
        r"""
        Append `text`::

            e = MTextEditor("First paragraph.\P")
            e += "Second paragraph.\P")

        """
        self.text += text
        return self

    def __str__(self) -> str:
        """Returns the MTEXT content attribute :attr:`text`."""
        return self.text

    def clear(self):
        """Reset the content to an empty string."""
        self.text = ""

    def font(
        self, name: str, bold: bool = False, italic: bool = False
    ) -> "MTextEditor":
        """Set the text font by the font family name. Changing the font height
        should be done by the :meth:`height` or the :meth:`scale_height` method.
        The font family name is the name shown in font selection widgets in
        desktop applications: "Arial", "Times New Roman", "Comic Sans MS".
        Switching the codepage is not supported.

        Args:
            name: font family name
            bold: flag
            italic: flag

        """
        # c0 = current codepage
        # The current implementation of ezdxf writes everything in one
        # encoding, defined by $DWGCODEPAGE < DXF R2007 or utf8 for DXF R2007+
        # Switching codepage makes no sense!
        # p0 = current text size
        # Text size should be changed by \H<factor>x;
        return self.append(rf"\f{name}|b{int(bold)}|i{int(italic)};")

    def scale_height(self, factor: float) -> "MTextEditor":
        """Scale the text height by a `factor`. This scaling will accumulate,
        which means starting at height 2.5 and scaling by 2 and again by 3 will
        set the text height to 2.5 x 2 x 3 = 15. The current text height is not
        stored in the :class:`MTextEditor`, you have to track the text height by
        yourself! The initial text height is stored in the
        :class:`~ezdxf.entities.MText` entity as DXF attribute
        :class:`~ezdxf.entities.MText.dxf.char_height`.

        """
        return self.append(rf"\H{round(factor, 3)}x;")

    def height(self, height: float) -> "MTextEditor":
        """Set the absolute text height in drawing units."""
        return self.append(rf"\H{round(height, 3)};")

    def width_factor(self, factor: float) -> "MTextEditor":
        """Set the absolute text width factor."""
        return self.append(rf"\W{round(factor, 3)};")

    def char_tracking_factor(self, factor: float) -> "MTextEditor":
        """Set the absolute character tracking factor."""
        return self.append(rf"\T{round(factor, 3)};")

    def oblique(self, angle: int) -> "MTextEditor":
        """Set the text oblique angle in degrees, vertical is 0, a value of 15
        will lean the text 15 degree to the right.

        """
        return self.append(rf"\Q{int(angle)};")

    def color(self, name: str) -> "MTextEditor":
        """Set the text color by color name: "red", "yellow", "green", "cyan",
        "blue", "magenta" or "white".

        """
        return self.aci(const.MTEXT_COLOR_INDEX[name.lower()])

    def aci(self, aci: int) -> "MTextEditor":
        """Set the text color by :ref:`ACI` in range [0, 256]."""
        if 0 <= aci <= 256:
            return self.append(rf"\C{aci};")
        else:
            raise ValueError("aci not in range [0, 256]")

    def rgb(self, rgb: RGB) -> "MTextEditor":
        """Set the text color as RGB value."""
        r, g, b = rgb
        return self.append(rf"\c{rgb2int((b, g, r))};")

    def stack(self, upr: str, lwr: str, t: str = "^") -> "MTextEditor":
        r"""Append stacked text `upr` over `lwr`, argument `t` defines the
        kind of stacking, the space " " after the "^" will be added
        automatically to avoid caret decoding:

        .. code-block:: none

            "^": vertical stacked without divider line, e.g. \SA^ B:
                 A
                 B

            "/": vertical stacked with divider line,  e.g. \SX/Y:
                 X
                 -
                 Y

            "#": diagonal stacked, with slanting divider line, e.g. \S1#4:
                 1/4

        """
        if t not in "^/#":
            raise ValueError(f"invalid type symbol: {t}")
        # space " " after "^" is required to avoid caret decoding
        if t == "^":
            t += " "
        return self.append(rf"\S{upr}{t}{lwr};")

    def group(self, text: str) -> "MTextEditor":
        """Group `text`, all properties changed inside a group are reverted at
        the end of the group. AutoCAD supports grouping up to 8 levels.

        """
        return self.append(f"{{{text}}}")

    def underline(self, text: str) -> "MTextEditor":
        """Append `text` with a line below the text."""
        return self.append(rf"\L{text}\l")

    def overline(self, text: str) -> "MTextEditor":
        """Append `text` with a line above the text."""
        return self.append(rf"\O{text}\o")

    def strike_through(self, text: str) -> "MTextEditor":
        """Append `text` with a line through the text."""
        return self.append(rf"\K{text}\k")

    def paragraph(self, props: ParagraphProperties) -> "MTextEditor":
        """Set paragraph properties by a :class:`ParagraphProperties` object."""
        return self.append(props.tostring())

    def bullet_list(
        self, indent: float, bullets: Iterable[str], content: Iterable[str]
    ) -> "MTextEditor":
        """Build bulleted lists by utilizing paragraph indentation and a
        tabulator stop. Any string can be used as bullet. Indentation is
        a multiple of the initial MTEXT char height (see also docs about
        :class:`ParagraphProperties`), which means indentation in
        drawing units is :attr:`MText.dxf.char_height` x `indent`.

        Useful UTF bullets:

        - "bull" U+2022 = • (Alt Numpad 7)
        - "circle" U+25CB = ○ (Alt Numpad 9)

        For numbered lists just use numbers as bullets::

            MTextEditor.bullet_list(
                indent=2,
                bullets=["1.", "2."],
                content=["first", "second"],
            )

        Args:
            indent: content indentation as multiple of the initial MTEXT char height
            bullets: iterable of bullet strings, e.g. :code:`["-"] * 3`,
                for 3 dashes as bullet strings
            content: iterable of list item strings, one string per list item,
                list items should not contain new line or new paragraph commands.

        """
        items = MTextEditor().paragraph(
            ParagraphProperties(
                indent=-indent * 0.75,  # like BricsCAD
                left=indent,
                tab_stops=(indent,),
            )
        )
        items.append(
            "".join(
                b + self.TAB + c + self.NEW_PARAGRAPH
                for b, c in zip(bullets, content)
            )
        )
        return self.group(str(items))


class MTextContext:
    """Internal class to store the MTEXT context state."""

    def __init__(self):
        self._stroke: int = 0
        self.continue_stroke: bool = False
        self._aci = 7  # used if rgb is None
        self.rgb: Optional[RGB] = None  # overrules aci
        self.align: MTextLineAlignment = MTextLineAlignment.BOTTOM
        self.font_face: FontFace = FontFace()  # is immutable
        self.cap_height: float = 1.0
        self.width_factor: float = 1.0
        self.char_tracking_factor: float = 1.0
        self.oblique: float = (
            0.0  # in degrees, where 0 is vertical (TEXT entity)
        )
        self.paragraph = ParagraphProperties()

    def __copy__(self) -> "MTextContext":
        p = MTextContext()
        p._stroke = self._stroke
        p.continue_stroke = self.continue_stroke
        p._aci = self._aci
        p.rgb = self.rgb
        p.align = self.align
        p.font_face = self.font_face  # is immutable
        p.cap_height = self.cap_height
        p.width_factor = self.width_factor
        p.char_tracking_factor = self.char_tracking_factor
        p.oblique = self.oblique
        p.paragraph = self.paragraph  # is immutable
        return p

    copy = __copy__

    def __hash__(self):
        return hash(
            (
                self._stroke,
                self.continue_stroke,
                self._aci,
                self.rgb,
                self.align,
                self.font_face,
                self.cap_height,
                self.width_factor,
                self.char_tracking_factor,
                self.oblique,
                self.paragraph,
            )
        )

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    @property
    def aci(self) -> int:
        return self._aci

    @aci.setter
    def aci(self, aci: int):
        if 0 <= aci <= 256:
            self._aci = aci
            self.rgb = None  # clear rgb
        else:
            raise ValueError("aci not in range[0,256]")

    def _set_stroke_state(
        self, stroke: MTextStroke, state: bool = True
    ) -> None:
        """Set/clear binary `stroke` flag in `self._stroke`.

        Args:
            stroke: set/clear stroke flag
            state: ``True`` for setting, ``False`` for clearing

        """
        if state:
            self._stroke |= stroke
        else:
            self._stroke &= ~stroke

    @property
    def underline(self) -> bool:
        return bool(self._stroke & MTextStroke.UNDERLINE)

    @underline.setter
    def underline(self, value: bool) -> None:
        self._set_stroke_state(MTextStroke.UNDERLINE, value)

    @property
    def strike_through(self) -> bool:
        return bool(self._stroke & MTextStroke.STRIKE_THROUGH)

    @strike_through.setter
    def strike_through(self, value: bool) -> None:
        self._set_stroke_state(MTextStroke.STRIKE_THROUGH, value)

    @property
    def overline(self) -> bool:
        return bool(self._stroke & MTextStroke.OVERLINE)

    @overline.setter
    def overline(self, value: bool) -> None:
        self._set_stroke_state(MTextStroke.OVERLINE, value)

    @property
    def has_any_stroke(self) -> bool:
        return bool(self._stroke)


class TextScanner:
    __slots__ = ("_text", "_text_len", "_index")

    def __init__(self, text: str):
        self._text = str(text)
        self._text_len = len(self._text)
        self._index = 0

    @property
    def is_empty(self) -> bool:
        return self._index >= self._text_len

    @property
    def has_data(self) -> bool:
        return self._index < self._text_len

    def get(self) -> str:
        try:
            char = self._text[self._index]
            self._index += 1
        except IndexError:
            return ""
        return char

    def consume(self, count: int = 1) -> None:
        if count < 1:
            raise ValueError(count)
        self._index += count

    def fast_consume(self, count: int = 1) -> None:
        """consume() without safety check"""
        self._index += count

    def peek(self, offset: int = 0) -> str:
        if offset < 0:
            raise ValueError(offset)
        try:
            return self._text[self._index + offset]
        except IndexError:
            return ""

    def fast_peek(self, offset: int = 0) -> str:
        """peek() without safety check"""
        try:
            return self._text[self._index + offset]
        except IndexError:
            return ""

    def find(self, char: str, escape=False) -> int:
        """Return the index of the next `char`. If `escape` is True, backslash
        escaped `chars` will be ignored e.g. "\\;;" index of the next ";" is 1
        if `escape` is False and 2 if `escape` is True. Returns -1 if `char`
        was not found.

        Args:
            char: single letter as string
            escape: ignore backslash escaped chars if True.

        """

        scanner = self.__class__(self._text[self._index :])
        while scanner.has_data:
            c = scanner.peek()
            if escape and c == "\\" and scanner.peek(1) == char:
                scanner.consume(2)
                continue
            if c == char:
                return self._index + scanner._index
            scanner.consume(1)
        return -1

    def substr(self, stop: int) -> str:
        """Returns the substring from the current location until index < stop."""
        if stop < self._index:
            raise IndexError(stop)
        return self._text[self._index : stop]

    def tail(self) -> str:
        """Returns the unprocessed part of the content."""
        return self._text[self._index :]


class TokenType(enum.IntEnum):
    NONE = 0
    WORD = 1  # data = str
    STACK = 2  # data = upr: str, lwr:str, type:str
    SPACE = 3  # data = None
    NBSP = 4  # data = None
    TABULATOR = 5  # data = None
    NEW_PARAGRAPH = 6  # data = None
    NEW_COLUMN = 7  # data = None
    WRAP_AT_DIMLINE = 8  # data = None


class MTextToken:
    __slots__ = ("type", "ctx", "data")

    def __init__(self, t: TokenType, ctx: MTextContext, data=None):
        self.type: TokenType = t
        self.ctx: MTextContext = ctx
        self.data = data


RE_FLOAT = re.compile(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
RE_FLOAT_X = re.compile(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?([x]?)")

CHAR_TO_ALIGN = {
    "l": MTextParagraphAlignment.LEFT,
    "r": MTextParagraphAlignment.RIGHT,
    "c": MTextParagraphAlignment.CENTER,
    "j": MTextParagraphAlignment.JUSTIFIED,
    "d": MTextParagraphAlignment.DISTRIBUTED,
}


class MTextParser:
    """Parses the MText content string and yields the content as tokens and
    the current MText properties as MTextContext object. The context object is
    treated internally as immutable object and should be treated by the client
    the same way.

    The parser works as iterator and yields MTextToken objects.

    Args:
        content: MText content string
        ctx: initial MText context

    """

    __slots__ = ("ctx", "scanner", "_ctx_stack", "_continue_stroke")

    def __init__(self, content: str, ctx: MTextContext = None):
        if ctx is None:
            ctx = MTextContext()
        self.ctx = ctx
        self.scanner = TextScanner(caret_decode(content))
        self._ctx_stack: List[MTextContext] = []
        self._continue_stroke = False

    def __iter__(self):
        return self.parse()

    def push_ctx(self) -> None:
        self._ctx_stack.append(self.ctx)

    def pop_ctx(self) -> None:
        if self._ctx_stack:
            self.ctx = self._ctx_stack.pop()

    def parse(self) -> Iterable[MTextToken]:
        # localize method calls
        scanner = self.scanner
        consume = scanner.fast_consume
        peek = scanner.fast_peek
        followup_token: Optional[TokenType] = None
        space_token = TokenType.SPACE
        word_token = TokenType.WORD

        def word_and_token(word, token):
            nonlocal followup_token
            consume()
            if word:
                followup_token = token
                return word_token, word
            else:
                return token, None

        def next_token():
            word = ""
            while scanner.has_data:
                escape = False
                letter = peek()
                if letter == "\\":
                    # known escape sequences: "\\", "\{", "\}"
                    if peek(1) in "\\{}":
                        escape = True
                        consume()  # leading backslash
                        letter = peek()
                    else:
                        # A non escaped backslash is always the end of a word.
                        if word:
                            # Do not consume backslash!
                            return word_token, word
                        consume()  # leading backslash
                        cmd = scanner.get()
                        if cmd == "~":
                            return TokenType.NBSP, None
                        if cmd == "P":
                            return TokenType.NEW_PARAGRAPH, None
                        if cmd == "N":
                            return TokenType.NEW_COLUMN, None
                        if cmd == "X":
                            return TokenType.WRAP_AT_DIMLINE, None
                        if cmd == "S":
                            return self.parse_stacking()
                        if cmd:
                            try:
                                self.parse_properties(cmd)
                            except ValueError:
                                # print invalid escaped letters verbatim
                                word += letter + cmd
                        continue

                # process control chars, caret decoding is already done!
                if letter < " ":
                    if letter == "\t":
                        return word_and_token(word, TokenType.TABULATOR)
                    elif letter == "\n":  # LF
                        return word_and_token(word, TokenType.NEW_PARAGRAPH)
                    else:  # replace other control chars by a space
                        letter = " "
                elif letter == "%" and peek(1) == "%":
                    code = peek(2).lower()
                    special_char = const.SPECIAL_CHAR_ENCODING.get(code)
                    if special_char:
                        consume(2)  # %%
                        letter = special_char

                if letter == " ":
                    return word_and_token(word, space_token)

                if not escape:
                    if letter == "{":
                        if word:
                            return word_token, word
                        else:
                            consume(1)
                            self.push_ctx()
                            continue
                    elif letter == "}":
                        if word:
                            return word_token, word
                        else:
                            consume(1)
                            self.pop_ctx()
                            continue

                # any unparsed unicode letter can be used in a word
                consume()
                if letter >= " ":
                    word += letter

            if word:
                return word_token, word
            else:
                return TokenType.NONE, None

        while True:
            type_, data = next_token()
            if type_:
                yield MTextToken(type_, self.ctx, data)
                if followup_token:
                    yield MTextToken(followup_token, self.ctx, None)
                    followup_token = None
            else:
                break

    def parse_stacking(self) -> Tuple:
        """Returns a tuple of strings: (numerator, denominator, type).
        The numerator and denominator is always a single and can contain spaces,
        which are not decoded as separate tokens. The type string is "^" for
        a limit style fraction without a line, "/" for a horizontal fraction
        and "#" for a diagonal fraction. If the expression does not contain
        any stacking type char, the type and denominator string are empty "".

        """

        def peek_char():
            c = stacking_scanner.peek()
            if ord(c) < 32:  # replace all control chars by space
                c = " "
            return c

        def get_next_char():
            escape = False
            c = peek_char()
            # escape sequences: remove backslash and return next char
            if c == "\\":
                escape = True
                stacking_scanner.consume(1)
                c = peek_char()
            stacking_scanner.consume(1)
            return c, escape

        def parse_numerator() -> Tuple[str, str]:
            word = ""
            while stacking_scanner.has_data:
                c, escape = get_next_char()
                if not escape and c in "^/#":  # scan until stacking type char
                    return word, c
                word += c
            return word, ""

        def parse_denominator() -> str:
            word = ""
            while stacking_scanner.has_data:
                word += get_next_char()[0]
            return word

        stacking_scanner = TextScanner(self.extract_expression(escape=True))
        numerator, stacking_type = parse_numerator()
        denominator = parse_denominator() if stacking_type else ""
        return TokenType.STACK, (numerator, denominator, stacking_type)

    def parse_properties(self, cmd: str) -> None:
        # Treat the existing context as immutable, create a new one:
        new_ctx = self.ctx.copy()
        if cmd == "L":
            new_ctx.underline = True
            self._continue_stroke = True
        elif cmd == "l":
            new_ctx.underline = False
            if not new_ctx.has_any_stroke:
                self._continue_stroke = False
        elif cmd == "O":
            new_ctx.overline = True
            self._continue_stroke = True
        elif cmd == "o":
            new_ctx.overline = False
            if not new_ctx.has_any_stroke:
                self._continue_stroke = False
        elif cmd == "K":
            new_ctx.strike_through = True
            self._continue_stroke = True
        elif cmd == "k":
            new_ctx.strike_through = False
            if not new_ctx.has_any_stroke:
                self._continue_stroke = False
        elif cmd == "A":
            self.parse_align(new_ctx)
        elif cmd == "C":
            self.parse_aci_color(new_ctx)
        elif cmd == "c":
            self.parse_rgb_color(new_ctx)
        elif cmd == "H":
            self.parse_height(new_ctx)
        elif cmd == "W":
            self.parse_width(new_ctx)
        elif cmd == "Q":
            self.parse_oblique(new_ctx)
        elif cmd == "T":
            self.parse_char_tracking(new_ctx)
        elif cmd == "p":
            self.parse_paragraph_properties(new_ctx)
        elif cmd == "f" or cmd == "F":
            self.parse_font_properties(new_ctx)
        else:  # unknown commands
            raise ValueError("unknown command")
        new_ctx.continue_stroke = self._continue_stroke
        self.ctx = new_ctx

    def parse_align(self, ctx: MTextContext):
        char = self.scanner.get()  # always consume next char
        if char in "012":
            ctx.align = MTextLineAlignment(int(char))
        else:
            ctx.align = MTextLineAlignment.BOTTOM
        self.consume_optional_terminator()

    def parse_height(self, ctx: MTextContext):
        ctx.cap_height = self.parse_float_value_or_factor(ctx.cap_height)
        self.consume_optional_terminator()

    def parse_width(self, ctx: MTextContext):
        ctx.width_factor = self.parse_float_value_or_factor(ctx.width_factor)
        self.consume_optional_terminator()

    def parse_char_tracking(self, ctx: MTextContext):
        ctx.char_tracking_factor = self.parse_float_value_or_factor(
            ctx.char_tracking_factor
        )
        self.consume_optional_terminator()

    def parse_float_value_or_factor(self, value) -> float:
        expr = self.extract_float_expression(relative=True)
        if expr:
            if expr.endswith("x"):
                factor = float(expr[:-1])
                value *= abs(factor)
            else:
                value = abs(float(expr))
        return value

    def parse_oblique(self, ctx: MTextContext):
        oblique_expr = self.extract_float_expression(relative=False)
        if oblique_expr:
            ctx.oblique = float(oblique_expr)
        self.consume_optional_terminator()

    def parse_aci_color(self, ctx: MTextContext):
        aci_expr = self.extract_int_expression()
        if aci_expr:
            aci = int(aci_expr)
            if aci < 257:
                ctx.aci = aci
                ctx.rgb = None
        self.consume_optional_terminator()

    def parse_rgb_color(self, ctx: MTextContext):
        rgb_expr = self.extract_int_expression()
        if rgb_expr:
            # in reversed order!
            b, g, r = int2rgb(int(rgb_expr) & 0xFFFFFF)
            ctx.rgb = r, g, b
        self.consume_optional_terminator()

    def extract_float_expression(self, relative=False) -> str:
        result = ""
        tail = self.scanner.tail()
        pattern = RE_FLOAT_X if relative else RE_FLOAT
        match = re.match(pattern, tail)
        if match:
            start, end = match.span()
            result = tail[start:end]
            self.scanner.consume(end)
        return result

    def extract_int_expression(self) -> str:
        result = ""
        tail = self.scanner.tail()
        match = re.match(r"\d+", tail)
        if match:
            start, end = match.span()
            result = tail[start:end]
            self.scanner.consume(end)
        return result

    def extract_expression(self, escape=False) -> str:
        """Returns the next expression from the current location until
        the terminating ";". The terminating semi-colon is not included.
        Skips escaped "\\;" semi-colons if `escape` is True.

        """
        stop = self.scanner.find(";", escape=escape)
        if stop < 0:  # ";" not found
            expr = self.scanner.tail()  # scan until end of content
        else:
            expr = self.scanner.substr(stop)  # exclude ";"
        # skip the expression in the main scanner
        self.scanner.consume(len(expr) + 1)  # include ";"
        return expr

    def parse_paragraph_properties(self, ctx: MTextContext):
        def parse_float() -> float:
            value = 0.0
            expr = parse_float_expr()
            if expr:
                value = float(expr)
            return value

        def parse_float_expr() -> str:
            expr = ""
            tail = paragraph_scanner.tail()
            match = re.match(RE_FLOAT, tail)
            if match:
                start, end = match.span()
                expr = tail[start:end]
                paragraph_scanner.consume(end)
                skip_commas()
            return expr

        def skip_commas():
            while paragraph_scanner.peek() == ",":
                paragraph_scanner.consume(1)

        paragraph_scanner = TextScanner(self.extract_expression())
        indent, left, right, align, tab_stops = ctx.paragraph  # NamedTuple
        while paragraph_scanner.has_data:
            cmd = paragraph_scanner.get()
            if cmd == "i":
                indent = parse_float()
            elif cmd == "l":
                left = parse_float()
            elif cmd == "r":
                right = parse_float()
            elif cmd == "x":
                pass  # ignore
            elif cmd == "q":
                adjustment = paragraph_scanner.get()
                align = CHAR_TO_ALIGN.get(
                    adjustment, MTextParagraphAlignment.DEFAULT
                )
                skip_commas()
            elif cmd == "t":
                tab_stops = []
                while paragraph_scanner.has_data:  # parse to end
                    type_ = paragraph_scanner.peek()
                    if type_ == "r" or type_ == "c":
                        paragraph_scanner.consume()
                        tab_stops.append(type_ + parse_float_expr())
                    else:
                        float_expr = parse_float_expr()
                        if float_expr:
                            tab_stops.append(float(float_expr))
                        else:
                            # invalid float expression, consume invalid letter
                            # and try again:
                            paragraph_scanner.consume()
        ctx.paragraph = ParagraphProperties(
            indent, left, right, align, tuple(tab_stops)
        )

    def parse_font_properties(self, ctx: MTextContext):
        parts = self.extract_expression().split("|")
        # an empty font family name does not change the font properties
        if parts and parts[0]:
            name = parts[0]
            style = "normal"
            weight = "normal"
            # ignore codepage and pitch - it seems not to be used in newer
            # CAD applications.
            for part in parts[1:]:
                if part.startswith("b1"):
                    weight = "bold"
                elif part.startswith("i1"):
                    style = "italic"
            ctx.font_face = FontFace(family=name, style=style, weight=weight)

    def consume_optional_terminator(self):
        if self.scanner.peek() == ";":
            self.scanner.consume(1)


def load_mtext_content(tags: "Tags") -> str:
    tail = ""
    content = ""
    for code, value in tags:
        if code == 1:
            tail = value
        elif code == 3:
            content += value
    return escape_dxf_line_endings(content + tail)


def has_inline_formatting_codes(text: str) -> bool:
    """Returns `True` if `text` contains any MTEXT inline formatting codes."""
    # Each inline formatting code starts with a backslash "\".
    # Remove all special chars starting with a "\" and test if any backslashes
    # remain. Escaped backslashes "\\" may return false positive,
    # but they are rare.
    # Replacing multiple strings at once by "re" is much slower,
    # see profiling/string_replace.py
    return "\\" in text.replace(  # line breaks
        r"\P", ""
    ).replace(  # non breaking spaces
        r"\~", ""
    )



import dataclasses
from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from ezdxf import disassemble


class LinePolicy(Enum):
    """
    Attributes:
        SOLID: draw all lines as solid regardless of the linetype style
        APPROXIMATE: use the closest approximation available to the
            backend for rendering styled lines
        ACCURATE: analyse and render styled lines as accurately as
            possible. This approach is slower and is not well suited
            to interactive applications.
    """

    SOLID = auto()
    APPROXIMATE = auto()
    ACCURATE = auto()


class ProxyGraphicPolicy(Enum):
    """The action to take when an entity with a proxy graphic is encountered

    .. note::

        To get proxy graphics support proxy graphics have to be loaded:
        Set the global option :attr:`ezdxf.options.load_proxy_graphics` to
        ``True``, which is the default value.

        This can not prevent drawing proxy graphic inside of blocks,
        because this is outside of the domain of the drawing add-on!

    Attributes:
        IGNORE: do not display proxy graphics (skip_entity will be called instead)
        SHOW: if the entity cannot be rendered directly (eg if not implemented)
            but a proxy is present: display the proxy
        PREFER: display proxy graphics even for entities where direct rendering
            is available
    """

    IGNORE = auto()
    SHOW = auto()
    PREFER = auto()


class HatchPolicy(Enum):
    """The action to take when a HATCH entity is encountered

    Attributes:
        IGNORE: do not show HATCH entities at all
        SHOW_OUTLINE: show only the outline of HATCH entities
        SHOW_SOLID: show HATCH entities but draw with solid fill
            regardless of the pattern
        SHOW_APPROXIMATE_PATTERN: show HATCH entities using the closest
            approximation available to the current backend

    """

    IGNORE = auto()
    SHOW_OUTLINE = auto()
    SHOW_SOLID = auto()
    SHOW_APPROXIMATE_PATTERN = auto()


@dataclass(frozen=True)
class Configuration:
    """Configuration options for the :mod:`drawing` add-on.

    Attributes:
        pdsize: the size to draw POINT entities (in drawing units)
            set to None to use the $PDSIZE value from the dxf document header

            ======= ====================================================
            0       5% of draw area height
            <0      Specifies a percentage of the viewport size
            >0      Specifies an absolute size
            None    use the $PDMODE value from the dxf document header
            ======= ====================================================

        pdmode: point styling mode (see POINT documentation)

            see :class:`~ezdxf.entities.Point` class documentation

        measurement: whether to use metric or imperial units

            ======= ======================================================
            0       use imperial units (in, ft, yd, ...)
            1       use metric units (ISO meters)
            None    use the $MEASUREMENT value from the dxf document header
            ======= ======================================================

        show_defpoints: whether to show or filter out POINT entities on the defpoints layer
        proxy_graphic_policy: the action to take when a proxy graphic is encountered
        line_policy: the method to use when drawing styled lines (eg dashed,
            dotted etc)
        hatch_policy: the method to use when drawing HATCH entities
        infinite_line_length: the length to use when drawing infinite lines
        lineweight_scaling:
            set to 0.0 for a constant minimal width the current result is
            correct, in SVG the line width is 0.7 points for 0.25mm as
            required, but it often looks too thick
        min_lineweight: the minimum line width in 1/300 inch, set to None for
            let the the backend choose.
        min_dash_length: the minimum length for a dash when drawing a styled line
            (default value is arbitrary)
        max_flattening_distance: Max flattening distance in drawing units
            see Path.flattening documentation.
            The backend implementation should calculate an appropriate value,
            like 1 screen- or paper pixel on the output medium, but converted
            into drawing units. Sets Path() approximation accuracy
        circle_approximation_count: Approximate a full circle by `n` segments, arcs
            have proportional less segments. Only used for approximation of arcs
            in banded polylines.
    """

    pdsize: Optional[int]
    pdmode: Optional[int]
    measurement: Optional[int]
    show_defpoints: bool
    proxy_graphic_policy: ProxyGraphicPolicy
    line_policy: LinePolicy
    hatch_policy: HatchPolicy
    infinite_line_length: float
    lineweight_scaling: float
    min_lineweight: Optional[float]
    min_dash_length: float
    max_flattening_distance: float
    circle_approximation_count: int

    @staticmethod
    def defaults() -> "Configuration":
        return Configuration(
            pdsize=1,
            pdmode=0,
            measurement=None,
            show_defpoints=False,
            proxy_graphic_policy=ProxyGraphicPolicy.SHOW,
            line_policy=LinePolicy.APPROXIMATE,
            hatch_policy=HatchPolicy.SHOW_APPROXIMATE_PATTERN,
            infinite_line_length=20,
            lineweight_scaling=1.0,
            min_lineweight=None,
            min_dash_length=0.1,
            max_flattening_distance=disassemble.Primitive.max_flattening_distance,
            circle_approximation_count=128,
        )

    def with_changes(self, **kwargs) -> "Configuration":
        params = dataclasses.asdict(self)
        for k, v in kwargs.items():
            params[k] = v
        return Configuration(**params)



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterator, Iterable, Union, cast, Dict, Tuple
from collections import Counter, OrderedDict
import logging

from ezdxf.lldxf.const import DXFStructureError, DXF2004, DXF2000, DXFKeyError
from ezdxf.entities.dxfclass import DXFClass
from ezdxf.entities.dxfentity import DXFEntity, DXFTagStorage

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Drawing

logger = logging.getLogger("ezdxf")

# name: cpp_class_name (2), app_name (3), flags(90), was_a_proxy (280),
# is_an_entity (281)
# Multiple entries for 'name' are possible and supported, ClassSection stores
# entries with key: (name, cpp_class_name).
# 0 <ctrl> CLASS
# 1 <str> MPOLYGON
# 2 <str> AcDbMPolygon
# 3 <str> "AcMPolygonObj15|Version(1.0.0.0) Product Desc: Object enabler for the AcDbMPolyg ... odesk.com"
# 90 <int> 3071, b101111111111
# 280 <int> 0
# 281 <int> 1
CLASS_DEFINITIONS = {
    "ACDBDICTIONARYWDFLT": [
        "AcDbDictionaryWithDefault",
        "ObjectDBX Classes",
        0,
        0,
        0,
    ],
    "SUN": ["AcDbSun", "SCENEOE", 1153, 0, 0],
    "DICTIONARYVAR": ["AcDbDictionaryVar", "ObjectDBX Classes", 0, 0, 0],
    "TABLESTYLE": ["AcDbTableStyle", "ObjectDBX Classes", 4095, 0, 0],
    "MATERIAL": ["AcDbMaterial", "ObjectDBX Classes", 1153, 0, 0],
    "VISUALSTYLE": ["AcDbVisualStyle", "ObjectDBX Classes", 4095, 0, 0],
    "SCALE": ["AcDbScale", "ObjectDBX Classes", 1153, 0, 0],
    "MLEADERSTYLE": ["AcDbMLeaderStyle", "ACDB_MLEADERSTYLE_CLASS", 4095, 0, 0],
    "MLEADER": ["AcDbMLeader", "ACDB_MLEADER_CLASS", 3071, 0, 1],
    "MPOLYGON": ["AcDbMPolygon", "AcMPolygonObj15", 1025, 0, 1],
    "CELLSTYLEMAP": ["AcDbCellStyleMap", "ObjectDBX Classes", 1152, 0, 0],
    "EXACXREFPANELOBJECT": ["ExAcXREFPanelObject", "EXAC_ESW", 1025, 0, 0],
    "NPOCOLLECTION": [
        "AcDbImpNonPersistentObjectsCollection",
        "ObjectDBX Classes",
        1153,
        0,
        0,
    ],
    "LAYER_INDEX": ["AcDbLayerIndex", "ObjectDBX Classes", 0, 0, 0],
    "SPATIAL_INDEX": ["AcDbSpatialIndex", "ObjectDBX Classes", 0, 0, 0],
    "IDBUFFER": ["AcDbIdBuffer", "ObjectDBX Classes", 0, 0, 0],
    "DIMASSOC": ["AcDbDimAssoc", "AcDbDimAssoc", 0, 0, 0],
    "ACDBSECTIONVIEWSTYLE": [
        "AcDbSectionViewStyle",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "ACDBDETAILVIEWSTYLE": [
        "AcDbDetailViewStyle",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "IMAGEDEF": ["AcDbRasterImageDef", "ISM", 0, 0, 0],
    "RASTERVARIABLES": ["AcDbRasterVariables", "ISM", 0, 0, 0],
    "IMAGEDEF_REACTOR": ["AcDbRasterImageDefReactor", "ISM", 1, 0, 0],
    "IMAGE": ["AcDbRasterImage", "ISM", 2175, 0, 1],
    "PDFDEFINITION": ["AcDbPdfDefinition", "ObjectDBX Classes", 1153, 0, 0],
    "PDFUNDERLAY": ["AcDbPdfReference", "ObjectDBX Classes", 4095, 0, 1],
    "DWFDEFINITION": ["AcDbDwfDefinition", "ObjectDBX Classes", 1153, 0, 0],
    "DWFUNDERLAY": ["AcDbDwfReference", "ObjectDBX Classes", 1153, 0, 1],
    "DGNDEFINITION": ["AcDbDgnDefinition", "ObjectDBX Classes", 1153, 0, 0],
    "DGNUNDERLAY": ["AcDbDgnReference", "ObjectDBX Classes", 1153, 0, 1],
    "MENTALRAYRENDERSETTINGS": [
        "AcDbMentalRayRenderSettings",
        "SCENEOE",
        1024,
        0,
        0,
    ],
    "ACDBPLACEHOLDER": ["AcDbPlaceHolder", "ObjectDBX Classes", 0, 0, 0],
    "LAYOUT": ["AcDbLayout", "ObjectDBX Classes", 0, 0, 0],
    "SURFACE": ["AcDbSurface", "ObjectDBX Classes", 4095, 0, 1],
    "EXTRUDEDSURFACE": ["AcDbExtrudedSurface", "ObjectDBX Classes", 4095, 0, 1],
    "LOFTEDSURFACE": ["AcDbLoftedSurface", "ObjectDBX Classes", 0, 0, 1],
    "REVOLVEDSURFACE": ["AcDbRevolvedSurface", "ObjectDBX Classes", 0, 0, 1],
    "SWEPTSURFACE": ["AcDbSweptSurface", "ObjectDBX Classes", 0, 0, 1],
    "PLANESURFACE": ["AcDbPlaneSurface", "ObjectDBX Classes", 4095, 0, 1],
    "NURBSSURFACE": ["AcDbNurbSurface", "ObjectDBX Classes", 4095, 0, 1],
    "ACDBASSOCEXTRUDEDSURFACEACTIONBODY": [
        "AcDbAssocExtrudedSurfaceActionBody",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "ACDBASSOCLOFTEDSURFACEACTIONBODY": [
        "AcDbAssocLoftedSurfaceActionBody",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "ACDBASSOCREVOLVEDSURFACEACTIONBODY": [
        "AcDbAssocRevolvedSurfaceActionBody",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "ACDBASSOCSWEPTSURFACEACTIONBODY": [
        "AcDbAssocSweptSurfaceActionBody",
        "ObjectDBX Classes",
        1025,
        0,
        0,
    ],
    "HELIX": ["AcDbHelix", "ObjectDBX Classes", 4095, 0, 1],
    "WIPEOUT": ["AcDbWipeout", "WipeOut", 127, 0, 1],
    "WIPEOUTVARIABLES": ["AcDbWipeoutVariables", "WipeOut", 0, 0, 0],
    "FIELDLIST": ["AcDbFieldList", "ObjectDBX Classes", 1152, 0, 0],
    "GEODATA": ["AcDbGeoData", "ObjectDBX Classes", 4095, 0, 0],
    "SORTENTSTABLE": ["AcDbSortentsTable", "ObjectDBX Classes", 0, 0, 0],
    "ACAD_TABLE": ["AcDbTable", "ObjectDBX Classes", 1025, 0, 1],
    "ARC_DIMENSION": ["AcDbArcDimension", "ObjectDBX Classes", 1025, 0, 1],
    "LARGE_RADIAL_DIMENSION": [
        "AcDbRadialDimensionLarge",
        "ObjectDBX Classes",
        1025,
        0,
        1,
    ],
}

REQ_R2000 = [
    "ACDBDICTIONARYWDFLT",
    "SUN",
    "VISUALSTYLE",
    "MATERIAL",
    "SCALE",
    "TABLESTYLE",
    "MLEADERSTYLE",
    "DICTIONARYVAR",
    "CELLSTYLEMAP",
    "MENTALRAYRENDERSETTINGS",
    "ACDBDETAILVIEWSTYLE",
    "ACDBSECTIONVIEWSTYLE",
    "RASTERVARIABLES",
    "ACDBPLACEHOLDER",
    "LAYOUT",
]

REQ_R2004 = [
    "ACDBDICTIONARYWDFLT",
    "SUN",
    "VISUALSTYLE",
    "MATERIAL",
    "SCALE",
    "TABLESTYLE",
    "MLEADERSTYLE",
    "DICTIONARYVAR",
    "CELLSTYLEMAP",
    "MENTALRAYRENDERSETTINGS",
    "ACDBDETAILVIEWSTYLE",
    "ACDBSECTIONVIEWSTYLE",
    "RASTERVARIABLES",
]

REQUIRED_CLASSES = {
    DXF2000: REQ_R2000,
    DXF2004: REQ_R2004,
}


class ClassesSection:
    def __init__(
        self, doc: "Drawing" = None, entities: Iterable[DXFEntity] = None
    ):
        # Multiple entries for 'name' possible -> key is (name, cpp_class_name)
        # DXFClasses are not stored in the entities database, because CLASS has
        # no handle.
        self.classes: Dict[Tuple[str, str], DXFClass] = OrderedDict()
        self.doc = doc
        if entities is not None:
            self.load(iter(entities))

    def __iter__(self) -> Iterable[DXFClass]:
        return (cls for cls in self.classes.values())

    def load(self, entities: Iterator[DXFEntity]) -> None:
        section_head = cast(DXFTagStorage, next(entities))

        if section_head.dxftype() != "SECTION" or section_head.base_class[
            1
        ] != (2, "CLASSES"):
            raise DXFStructureError(
                "Critical structure error in CLASSES section."
            )

        for cls_entity in entities:
            if isinstance(cls_entity, DXFClass):
                self.register(cls_entity)
            else:
                logger.warning(
                    f"Ignored invalid DXF entity type '{cls_entity.dxftype()}'"
                    f" in section CLASSES."
                )

    def register(
        self, classes: Union[DXFClass, Iterable[DXFClass]] = None
    ) -> None:
        if classes is None:
            return

        if isinstance(classes, DXFClass):
            classes = (classes,)

        for dxfclass in classes:
            key = dxfclass.key
            if key not in self.classes:
                self.classes[key] = dxfclass

    def add_class(self, name: str):
        """Register a known class by `name`."""
        if name not in CLASS_DEFINITIONS:
            return
        cls_data = CLASS_DEFINITIONS[name]
        cls = DXFClass.new(doc=self.doc)
        cpp, app, flags, proxy, entity = cls_data
        cls.update_dxf_attribs(
            {
                "name": name,
                "cpp_class_name": cpp,
                "app_name": app,
                "flags": flags,
                "was_a_proxy": proxy,
                "is_an_entity": entity,
            }
        )
        self.register(cls)

    def get(self, name: str) -> DXFClass:
        """Returns the first class matching `name`.

        Storage key is the ``(name, cpp_class_name)`` tuple, because there are
        some classes with the same :attr:`name` but different
        :attr:`cpp_class_names`.

        """
        for cls in self.classes.values():
            if cls.dxf.name == name:
                return cls
        raise DXFKeyError(name)

    def add_required_classes(self, dxfversion: str) -> None:
        """Add all required CLASS definitions for `dxfversion`."""
        names = REQUIRED_CLASSES.get(dxfversion, REQ_R2004)
        for name in names:
            self.add_class(name)

        if self.doc is None:  # testing environment SUT
            return

        dxf_types_in_use = self.doc.entitydb.dxf_types_in_use()
        if "IMAGE" in dxf_types_in_use:
            self.add_class("IMAGE")
            self.add_class("IMAGEDEF")
            self.add_class("IMAGEDEF_REACTOR")
        if "PDFUNDERLAY" in dxf_types_in_use:
            self.add_class("PDFDEFINITION")
            self.add_class("PDFUNDERLAY")
        if "DWFUNDERLAY" in dxf_types_in_use:
            self.add_class("DWFDEFINITION")
            self.add_class("DWFUNDERLAY")
        if "DGNUNDERLAY" in dxf_types_in_use:
            self.add_class("DGNDEFINITION")
            self.add_class("DGNUNDERLAY")
        if "EXTRUDEDSURFACE" in dxf_types_in_use:
            self.add_class("EXTRUDEDSURFACE")
            self.add_class("ACDBASSOCEXTRUDEDSURFACEACTIONBODY")
        if "LOFTEDSURFACE" in dxf_types_in_use:
            self.add_class("LOFTEDSURFACE")
            self.add_class("ACDBASSOCLOFTEDSURFACEACTIONBODY")
        if "REVOLVEDSURFACE" in dxf_types_in_use:
            self.add_class("REVOLVEDSURFACE")
            self.add_class("ACDBASSOCREVOLVEDSURFACEACTIONBODY")
        if "SWEPTSURFACE" in dxf_types_in_use:
            self.add_class("SWEPTSURFACE")
            self.add_class("ACDBASSOCSWEPTSURFACEACTIONBODY")

        for dxftype in dxf_types_in_use:
            self.add_class(dxftype)

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        """Export DXF tags. (internal API)"""
        tagwriter.write_str("  0\nSECTION\n  2\nCLASSES\n")
        for dxfclass in self.classes.values():
            dxfclass.export_dxf(tagwriter)
        tagwriter.write_str("  0\nENDSEC\n")

    def update_instance_counters(self) -> None:
        """Update CLASS instance counter for all registered classes, requires
        DXF R2004+.
        """
        assert self.doc is not None
        if self.doc.dxfversion < DXF2004:
            return  # instance counter not supported
        counter: Dict[str, int] = Counter()
        # count all entities in the entity database
        for entity in self.doc.entitydb.values():
            counter[entity.dxftype()] += 1

        for dxfclass in self.classes.values():
            dxfclass.dxf.instance_count = counter[dxfclass.dxf.name]



# Copyright (c) 2020-2021, Matthew Broadway
# License: MIT License
from typing import List

from ezdxf.addons.drawing.backend import BackendInterface
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.math import Vec3


def draw_rect(points: List[Vec3], color: Color, out: BackendInterface):
    from ezdxf.addons.drawing import Properties

    props = Properties()
    props.color = color
    for a, b in zip(points, points[1:]):
        out.draw_line(a, b, props)



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Iterable
from ezdxf.lldxf import validator
from ezdxf.lldxf import const
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.types import DXFTag, DXFVertex
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import (
    DXF12,
    SUBCLASS_MARKER,
    DXFStructureError,
    DXFValueError,
    DXFTableEntryError,
)
from ezdxf.math import Vec3, Vec2, NULLVEC, X_AXIS, Y_AXIS
from ezdxf.tools import set_flag_state
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Drawing, DXFEntity

__all__ = ["Viewport"]

acdb_viewport = DefSubclass(
    "AcDbViewport",
    {
        # DXF reference: Center point (in WCS)
        # Correction to the DXF reference:
        # This point represents the center of the viewport in paper space units
        # (DCS), but is stored as 3D point inclusive z-axis!
        "center": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Width in paper space units:
        "width": DXFAttr(40, default=1),
        # Height in paper space units:
        "height": DXFAttr(41, default=1),
        # Viewport status field:
        # -1 = On, but is fully off screen, or is one of the viewports that is not
        #      active because the $MAXACTVP count is currently being exceeded.
        #  0 = Off
        # <positive value> = On and active. The value indicates the order of
        # stacking for the viewports, where 1 is the "active viewport", 2 is the
        # next, and so on:
        "status": DXFAttr(68, default=0),
        # Special VIEWPORT id == 1, this viewport defines the area of the layout
        # which is currently shown in the layout tab by the CAD application.
        # I guess this is meant by "active viewport" and therefore it is most likely
        # that this id is always 1.
        # This "active viewport" is mandatory for a valid DXF file.
        # BricsCAD set id to -1 if the viewport is off and 'status' (group code 68)
        # is not present.
        "id": DXFAttr(69, default=2),
        # DXF reference: View center point (in DCS):
        # Correction to the DXF reference:
        # This point represents the center point in model space (WCS) stored as
        # 2D point!
        "view_center_point": DXFAttr(12, xtype=XType.point2d, default=NULLVEC),
        "snap_base_point": DXFAttr(13, xtype=XType.point2d, default=NULLVEC),
        "snap_spacing": DXFAttr(14, xtype=XType.point2d, default=Vec2(10, 10)),
        "grid_spacing": DXFAttr(15, xtype=XType.point2d, default=Vec2(10, 10)),
        # View direction vector (WCS):
        "view_direction_vector": DXFAttr(
            16, xtype=XType.point3d, default=NULLVEC
        ),
        # View target point (in WCS):
        "view_target_point": DXFAttr(17, xtype=XType.point3d, default=NULLVEC),
        "perspective_lens_length": DXFAttr(42, default=50),
        "front_clip_plane_z_value": DXFAttr(43, default=0),
        "back_clip_plane_z_value": DXFAttr(44, default=0),
        # View height (in model space units):
        "view_height": DXFAttr(45, default=1),
        "snap_angle": DXFAttr(50, default=0),
        "view_twist_angle": DXFAttr(51, default=0),
        "circle_zoom": DXFAttr(72, default=100),
        # 331: Frozen layer object ID/handle (multiple entries may exist) (optional)
        # Viewport status bit-coded flags:
        # 1 (0x1) = Enables perspective mode
        # 2 (0x2) = Enables front clipping
        # 4 (0x4) = Enables back clipping
        # 8 (0x8) = Enables UCS follow
        # 16 (0x10) = Enables front clip not at eye
        # 32 (0x20) = Enables UCS icon visibility
        # 64 (0x40) = Enables UCS icon at origin
        # 128 (0x80) = Enables fast zoom
        # 256 (0x100) = Enables snap mode
        # 512 (0x200) = Enables grid mode
        # 1024 (0x400) = Enables isometric snap style
        # 2048 (0x800) = Enables hide plot mode
        # 4096 (0x1000) = kIsoPairTop. If set and kIsoPairRight is not set, then
        #   isopair top is enabled. If both kIsoPairTop and kIsoPairRight are set,
        #   then isopair left is enabled
        # 8192 (0x2000) = kIsoPairRight. If set and kIsoPairTop is not set, then
        #   isopair right is enabled
        # 16384 (0x4000) = Enables viewport zoom locking
        # 32768 (0x8000) = Currently always enabled
        # 65536 (0x10000) = Enables non-rectangular clipping
        # 131072 (0x20000) = Turns the viewport off
        # 262144 (0x40000) = Enables the display of the grid beyond the drawing
        #   limits
        # 524288 (0x80000) = Enable adaptive grid display
        # 1048576 (0x100000) = Enables subdivision of the grid below the set grid
        #   spacing when the grid display is adaptive
        # 2097152 (0x200000) = Enables grid follows workplane switching
        "flags": DXFAttr(90, default=0),
        # Clipping viewports: the following handle point to a graphical entity
        # located in the paperspace. Known supported entities:
        # LWPOLYLINE (2D POLYLINE), CIRCLE, ELLIPSE, closed SPLINE
        # Extract bounding- or clipping path: ezdxf.render.make_path()
        "clipping_boundary_handle": DXFAttr(340, default="0", optional=True),
        # Plot style sheet name assigned to this viewport
        "plot_style_name": DXFAttr(1, default=""),
        # Render mode:
        # 0 = 2D Optimized (classic 2D)
        # 1 = Wireframe
        # 2 = Hidden line
        # 3 = Flat shaded
        # 4 = Gouraud shaded
        # 5 = Flat shaded with wireframe
        # 6 = Gouraud shaded with wireframe
        "render_mode": DXFAttr(
            281,
            default=0,
            validator=validator.is_in_integer_range(0, 7),
            fixer=RETURN_DEFAULT,
        ),
        "ucs_per_viewport": DXFAttr(
            71,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "ucs_icon": DXFAttr(74, default=0),
        "ucs_origin": DXFAttr(110, xtype=XType.point3d, default=NULLVEC),
        "ucs_x_axis": DXFAttr(
            111,
            xtype=XType.point3d,
            default=X_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        "ucs_y_axis": DXFAttr(
            112,
            xtype=XType.point3d,
            default=Y_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Handle of AcDbUCSTableRecord if UCS is a named UCS.
        # If not present, then UCS is unnamed:
        "ucs_handle": DXFAttr(345),
        # Handle of AcDbUCSTableRecord of base UCS if UCS is orthographic (79 code
        # is non-zero). If not present and 79 code is non-zero, then base UCS is
        # taken to be WORLD:
        "ucs_base_handle": DXFAttr(346, optional=True),
        # UCS ortho type:
        # 0 = not orthographic
        # 1 = Top
        # 2 = Bottom
        # 3 = Front
        # 4 = Back
        # 5 = Left
        # 6 = Right
        "ucs_ortho_type": DXFAttr(
            79,
            default=0,
            validator=validator.is_in_integer_range(0, 7),
            fixer=RETURN_DEFAULT,
        ),
        "elevation": DXFAttr(146, default=0),
        # Shade plot mode:
        # 0 = As Displayed
        # 1 = Wireframe
        # 2 = Hidden
        # 3 = Rendered
        "shade_plot_mode": DXFAttr(
            170,
            dxfversion="AC1018",
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        ),
        # Frequency of major grid lines compared to minor grid lines
        "grid_frequency": DXFAttr(61, dxfversion="AC1021"),
        "background_handle": DXFAttr(332, dxfversion="AC1021"),
        "shade_plot_handle": DXFAttr(333, dxfversion="AC1021"),
        "visual_style_handle": DXFAttr(348, dxfversion="AC1021"),
        "default_lighting_flag": DXFAttr(
            292, dxfversion="AC1021", default=1, optional=True
        ),
        # Default lighting type:
        # 0 = One distant light
        # 1 = Two distant lights
        "default_lighting_type": DXFAttr(
            282,
            default=0,
            dxfversion="AC1021",
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "view_brightness": DXFAttr(141, dxfversion="AC1021"),
        "view_contrast": DXFAttr(142, dxfversion="AC1021"),
        # as AutoCAD Color Index
        "ambient_light_color_1": DXFAttr(
            63,
            dxfversion="AC1021",
            validator=validator.is_valid_aci_color,
        ),
        # as True Color:
        "ambient_light_color_2": DXFAttr(421, dxfversion="AC1021"),
        # as True Color:
        "ambient_light_color_3": DXFAttr(431, dxfversion="AC1021"),
        "sun_handle": DXFAttr(361, dxfversion="AC1021"),
        "ref_vp_object_1": DXFAttr(335, dxfversion="AC1021"),
        # unknown meaning, don't ask mozman
        "ref_vp_object_2": DXFAttr(343, dxfversion="AC1021"),
        # unknown meaning, don't ask mozman
        "ref_vp_object_3": DXFAttr(344, dxfversion="AC1021"),
        # unknown meaning, don't ask mozman
        "ref_vp_object_4": DXFAttr(91, dxfversion="AC1021"),
        # unknown meaning, don't ask mozman
    },
)
acdb_viewport_group_codes = group_code_mapping(acdb_viewport)
# Note:
# The ZOOM XP factor is calculated with the following formula:
# group_41 / group_45 (or pspace_height / mspace_height).

FROZEN_LAYER_GROUP_CODE = 331


@register_entity
class Viewport(DXFGraphic):
    """DXF VIEWPORT entity"""

    DXFTYPE = "VIEWPORT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_viewport)
    viewport_id = 2

    # Notes to viewport_id:
    # The id of the first viewport has to be 1, which is the definition of
    # paper space. For the following viewports it seems only important, that
    # the id is greater than 1.

    def get_next_viewport_id(self) -> int:
        current_id = Viewport.viewport_id
        Viewport.viewport_id += 1
        return current_id

    def __init__(self):
        super().__init__()
        self._frozen_layers: List[str] = []

    def _copy_data(self, entity: "DXFEntity") -> None:
        assert isinstance(entity, Viewport)
        entity._frozen_layers = self._frozen_layers

    @property
    def frozen_layers(self) -> List[str]:
        """Set/get frozen layers as list of layer names."""
        return self._frozen_layers

    @frozen_layers.setter
    def frozen_layers(self, names: Iterable[str]):
        self._frozen_layers = list(names)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.fast_load_dxfattribs(
                dxf, acdb_viewport_group_codes, subclass=2, log=False
            )
            if processor.r12:
                self.load_xdata_into_dxf_namespace()
            else:
                if len(tags):
                    tags = self.load_frozen_layer_handles(tags)
                if len(tags):
                    processor.log_unprocessed_tags(
                        tags, subclass=acdb_viewport.name
                    )
        return dxf

    def post_load_hook(self, doc: "Drawing"):
        super().post_load_hook(doc)
        bag: List[str] = []
        db = doc.entitydb
        for handle in self._frozen_layers:
            try:
                bag.append(db[handle].dxf.name)
            except KeyError:  # ignore non-existing layers
                pass
        self._frozen_layers = bag

    def load_frozen_layer_handles(self, tags: Tags) -> Tags:
        unprocessed_tags = Tags()
        for tag in tags:
            if tag.code == FROZEN_LAYER_GROUP_CODE:
                self._frozen_layers.append(tag.value)
            else:
                unprocessed_tags.append(tag)
        return unprocessed_tags

    def load_xdata_into_dxf_namespace(self) -> None:
        try:
            tags = [v for c, v in self.xdata.get_xlist("ACAD", "MVIEW")]  # type: ignore
        except DXFValueError:
            return
        tags = tags[3:-2]
        dxf = self.dxf
        flags = 0
        flags = set_flag_state(flags, const.VSF_FAST_ZOOM, bool(tags[11]))
        flags = set_flag_state(flags, const.VSF_SNAP_MODE, bool(tags[13]))
        flags = set_flag_state(flags, const.VSF_GRID_MODE, bool(tags[14]))
        flags = set_flag_state(
            flags, const.VSF_ISOMETRIC_SNAP_STYLE, bool(tags[15])
        )
        flags = set_flag_state(flags, const.VSF_HIDE_PLOT_MODE, bool(tags[24]))
        try:
            dxf.view_target_point = tags[0]
            dxf.view_direction_vector = tags[1]
            dxf.view_twist_angle = tags[2]
            dxf.view_height = tags[3]
            dxf.view_center_point = tags[4], tags[5]
            dxf.perspective_lens_length = tags[6]
            dxf.front_clip_plane_z_value = tags[7]
            dxf.back_clip_plane_z_value = tags[8]
            dxf.render_mode = tags[9]  # view_mode == render_mode ?
            dxf.circle_zoom = tags[10]
            # fast zoom flag : tag[11]
            dxf.ucs_icon = tags[12]
            # snap mode flag  = tags[13]
            # grid mode flag = tags[14]
            # isometric snap style = tags[15]
            # dxf.snap_isopair = tags[16] ???
            dxf.snap_angle = tags[17]
            dxf.snap_base_point = tags[18], tags[19]
            dxf.snap_spacing = tags[20], tags[21]
            dxf.grid_spacing = tags[22], tags[23]
            # hide plot flag  = tags[24]
        except IndexError:  # internal exception
            raise DXFStructureError("Invalid viewport entity - missing data")
        dxf.flags = flags
        self._frozen_layers = tags[26:]
        self.xdata.discard("ACAD")  # type: ignore

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion == DXF12:
            self.export_acdb_viewport_r12(tagwriter)
        else:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_viewport.name)
            self.dxf.export_dxf_attribs(
                tagwriter,
                [
                    "center",
                    "width",
                    "height",
                    "status",
                    "id",
                    "view_center_point",
                    "snap_base_point",
                    "snap_spacing",
                    "grid_spacing",
                    "view_direction_vector",
                    "view_target_point",
                    "perspective_lens_length",
                    "front_clip_plane_z_value",
                    "back_clip_plane_z_value",
                    "view_height",
                    "snap_angle",
                    "view_twist_angle",
                    "circle_zoom",
                ],
            )
            if len(self.frozen_layers):
                assert self.doc is not None, "valid DXF document required"
                layers = self.doc.layers
                for layer_name in self.frozen_layers:
                    try:
                        layer = layers.get(layer_name)
                    except DXFTableEntryError:
                        pass
                    else:
                        tagwriter.write_tag2(
                            FROZEN_LAYER_GROUP_CODE, layer.dxf.handle
                        )

            self.dxf.export_dxf_attribs(
                tagwriter,
                [
                    "flags",
                    "clipping_boundary_handle",
                    "plot_style_name",
                    "render_mode",
                    "ucs_per_viewport",
                    "ucs_icon",
                    "ucs_origin",
                    "ucs_x_axis",
                    "ucs_y_axis",
                    "ucs_handle",
                    "ucs_base_handle",
                    "ucs_ortho_type",
                    "elevation",
                    "shade_plot_mode",
                    "grid_frequency",
                    "background_handle",
                    "shade_plot_handle",
                    "visual_style_handle",
                    "default_lighting_flag",
                    "default_lighting_type",
                    "view_brightness",
                    "view_contrast",
                    "ambient_light_color_1",
                    "ambient_light_color_2",
                    "ambient_light_color_3",
                    "sun_handle",
                    "ref_vp_object_1",
                    "ref_vp_object_2",
                    "ref_vp_object_3",
                    "ref_vp_object_4",
                ],
            )

    def export_acdb_viewport_r12(self, tagwriter: "TagWriter"):
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "center",
                "width",
                "height",
                "status",
                "id",
            ],
        )
        tagwriter.write_tags(self.dxftags())

    def dxftags(self) -> Tags:
        def flag(flag):
            return 1 if self.dxf.flags & flag else 0

        dxf = self.dxf
        tags = [
            DXFTag(1001, "ACAD"),
            DXFTag(1000, "MVIEW"),
            DXFTag(1002, "{"),
            DXFTag(1070, 16),  # extended data version, always 16 for R11/12
            DXFVertex(1010, dxf.view_target_point),
            DXFVertex(1010, dxf.view_direction_vector),
            DXFTag(1040, dxf.view_twist_angle),
            DXFTag(1040, dxf.view_height),
            DXFTag(1040, dxf.view_center_point[0]),
            DXFTag(
                1040,
                dxf.view_center_point[1],
            ),
            DXFTag(1040, dxf.perspective_lens_length),
            DXFTag(1040, dxf.front_clip_plane_z_value),
            DXFTag(1040, dxf.back_clip_plane_z_value),
            DXFTag(1070, dxf.render_mode),
            DXFTag(1070, dxf.circle_zoom),
            DXFTag(1070, flag(const.VSF_FAST_ZOOM)),
            DXFTag(1070, dxf.ucs_icon),
            DXFTag(1070, flag(const.VSF_SNAP_MODE)),
            DXFTag(1070, flag(const.VSF_GRID_MODE)),
            DXFTag(1070, flag(const.VSF_ISOMETRIC_SNAP_STYLE)),
            DXFTag(1070, 0),  # snap isopair ???
            DXFTag(1040, dxf.snap_angle),
            DXFTag(1040, dxf.snap_base_point[0]),
            DXFTag(1040, dxf.snap_base_point[1]),
            DXFTag(1040, dxf.snap_spacing[0]),
            DXFTag(1040, dxf.snap_spacing[1]),
            DXFTag(1040, dxf.grid_spacing[0]),
            DXFTag(1040, dxf.grid_spacing[1]),
            DXFTag(1070, flag(const.VSF_HIDE_PLOT_MODE)),
            DXFTag(1002, "{"),  # start frozen layer list
        ]
        tags.extend(
            DXFTag(1003, layer_name) for layer_name in self.frozen_layers
        )
        tags.extend(
            [
                DXFTag(1002, "}"),  # end of frozen layer list
                DXFTag(1002, "}"),  # MVIEW
            ]
        )
        return Tags(tags)

    def rename_frozen_layer(self, old_name: str, new_name: str) -> None:
        assert self.doc is not None, "valid DXF document required"
        key = self.doc.layers.key
        old_key = key(old_name)
        self.frozen_layers = [
            (name if key(name) != old_key else new_name)
            for name in self.frozen_layers
        ]

    def boundary_path(self) -> List[Vec3]:
        center = self.dxf.center
        cx = center.x
        cy = center.y
        width2 = self.dxf.width / 2
        height2 = self.dxf.height / 2
        # TODO: clipping path support for the Viewport entity
        return [
            Vec3(cx - width2, cy - height2),
            Vec3(cx + width2, cy - height2),
            Vec3(cx + width2, cy + height2),
            Vec3(cx - width2, cy + height2),
        ]

    def has_clipping_path(self) -> bool:
        _flag = self.dxf.flags & const.VSF_NON_RECTANGULAR_CLIPPING
        if _flag:
            handle = self.dxf.clipping_boundary_handle
            return handle != "0"
        return False



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Tuple, Optional, List, Iterator
import abc

from ezdxf.math import Vec3, Vec2

if TYPE_CHECKING:
    from ezdxf.math import Vertex, AnyVec

__all__ = ["BoundingBox2d", "BoundingBox"]


class AbstractBoundingBox:
    __slots__ = ("extmin", "extmax")

    def __init__(self, vertices: Iterable["Vertex"] = None):
        self.extmax: Optional["AnyVec"] = None
        self.extmin: Optional["AnyVec"] = None
        if vertices is not None:
            try:
                self.extmin, self.extmax = self.extends_detector(vertices)
            except ValueError:
                # No or invalid data creates an empty BoundingBox
                pass

    def __str__(self) -> str:
        return f"[{self.extmin}, {self.extmax}]"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        if self.has_data:
            return f"{name}({self.__str__()})"
        else:
            return f"{name}()"

    def __iter__(self) -> Iterator["AnyVec"]:
        if self.has_data:
            yield self.extmin
            yield self.extmax

    @abc.abstractmethod
    def extends_detector(
        self, vertices: Iterable["Vertex"]
    ) -> Tuple["AnyVec", "AnyVec"]:
        pass

    @abc.abstractmethod
    def inside(self, vertex: "Vertex") -> bool:
        pass

    @abc.abstractmethod
    def intersect(self, other: "AbstractBoundingBox") -> bool:
        pass

    def any_inside(self, vertices: Iterable["Vertex"]) -> bool:
        """Returns ``True`` if any vertex is inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.has_data:
            return any(self.inside(v) for v in vertices)
        return False

    def all_inside(self, vertices: Iterable["Vertex"]) -> bool:
        """Returns ``True`` if all vertices are inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.has_data:
            # all() returns True for an empty set of vertices
            has_any = False
            for v in vertices:
                has_any = True
                if not self.inside(v):
                    return False
            return has_any
        return False

    @property
    def has_data(self) -> bool:
        """Returns ``True`` if bonding box is not empty"""
        return self.extmin is not None

    @property
    def is_empty(self) -> bool:
        """Returns ``True`` if bonding box is empty"""
        return self.extmin is None

    @property
    def size(self):
        """Returns size of bounding box."""
        return self.extmax - self.extmin

    @property
    def center(self):
        """Returns center of bounding box."""
        return self.extmin.lerp(self.extmax)

    def extend(self, vertices: Iterable["Vertex"]) -> None:
        """Extend bounds by `vertices`.

        Args:
            vertices: iterable of Vertex objects

        """
        v = list(vertices)
        if not v:
            return
        if self.has_data:
            v.extend([self.extmin, self.extmax])
        self.extmin, self.extmax = self.extends_detector(v)

    def union(self, other: "AbstractBoundingBox"):
        """Returns a new bounding box as union of this and `other` bounding
        box.
        """
        vertices: List["AnyVec"] = []
        if self.has_data:
            vertices.extend(self)
        if other.has_data:
            vertices.extend(other)
        return self.__class__(vertices)

    def rect_vertices(self) -> Tuple[Vec2, ...]:
        """Returns the corners of the bounding box in the xy-plane as
        :class:`Vec2` objects.
        """
        if self.has_data:  # extmin is not None!
            x0, y0, *_ = self.extmin  # type: ignore
            x1, y1, *_ = self.extmax  # type: ignore
            return Vec2(x0, y0), Vec2(x1, y0), Vec2(x1, y1), Vec2(x0, y1)
        else:
            raise ValueError("empty bounding box")


class BoundingBox(AbstractBoundingBox):
    """3D bounding box.

    Args:
        vertices: iterable of ``(x, y, z)`` tuples or :class:`Vec3` objects

    """

    def extends_detector(
        self, vertices: Iterable["Vertex"]
    ) -> Tuple[Vec3, Vec3]:
        return extends3d(vertices)

    def inside(self, vertex: "Vertex") -> bool:
        """Returns ``True`` if `vertex` is inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.extmin is None or self.extmax is None:
            return False
        x, y, z = Vec3(vertex).xyz
        xmin, ymin, zmin = self.extmin.xyz
        xmax, ymax, zmax = self.extmax.xyz
        return (
            (xmin <= x <= xmax) and (ymin <= y <= ymax) and (zmin <= z <= zmax)
        )

    def intersect(self, other: "AbstractBoundingBox") -> bool:
        """Returns `True` if this bounding box intersects with `other`.

        Touching bounding boxes do not intersect!

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        # Check for a separating axis:
        if (
            self.extmin is None
            or self.extmax is None
            or other.extmin is None
            or other.extmax is None
        ):
            return False

        # Check for a separating axis:
        if self.extmin.x >= other.extmax.x:
            return False
        if self.extmax.x <= other.extmin.x:
            return False
        if self.extmin.y >= other.extmax.y:
            return False
        if self.extmax.y <= other.extmin.y:
            return False
        if self.extmin.z >= other.extmax.z:
            return False
        if self.extmax.z <= other.extmin.z:
            return False
        return True

    def cube_vertices(self) -> Tuple[Vec3, ...]:
        """Returns the 3D corners of the bounding box as :class:`Vec3` objects."""
        if self.extmin is not None and self.extmax is not None:
            x0, y0, z0 = self.extmin
            x1, y1, z1 = self.extmax
            return (
                Vec3(x0, y0, z0),
                Vec3(x1, y0, z0),
                Vec3(x1, y1, z0),
                Vec3(x0, y1, z0),
                Vec3(x0, y0, z1),
                Vec3(x1, y0, z1),
                Vec3(x1, y1, z1),
                Vec3(x0, y1, z1),
            )
        else:
            raise ValueError("empty bounding box")


class BoundingBox2d(AbstractBoundingBox):
    """Optimized 2D bounding box.

    Args:
        vertices: iterable of ``(x, y[, z])`` tuples or :class:`Vec3` objects

    """

    def extends_detector(
        self, vertices: Iterable["Vertex"]
    ) -> Tuple[Vec2, Vec2]:
        return extends2d(vertices)

    def inside(self, vertex: "Vertex") -> bool:
        """Returns ``True`` if `vertex` is inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.extmin is None or self.extmax is None:
            return False
        v = Vec2(vertex)
        min_ = self.extmin
        max_ = self.extmax
        return (min_.x <= v.x <= max_.x) and (min_.y <= v.y <= max_.y)

    def intersect(self, other: "AbstractBoundingBox") -> bool:
        """Returns `True` if this bounding box intersects with `other`.

        Touching bounding boxes do not intersect!

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        if (
            self.extmin is None
            or self.extmax is None
            or other.extmin is None
            or other.extmax is None
        ):
            return False
        # Check for a separating axis:
        if self.extmin.x >= other.extmax.x:
            return False
        if self.extmax.x <= other.extmin.x:
            return False
        if self.extmin.y >= other.extmax.y:
            return False
        if self.extmax.y <= other.extmin.y:
            return False
        return True


def extends3d(vertices: Iterable["Vertex"]) -> Tuple[Vec3, Vec3]:
    minx, miny, minz = None, None, None
    maxx, maxy, maxz = None, None, None
    for v in vertices:
        v = Vec3(v)
        if minx is None:
            minx, miny, minz = v.xyz  # type: ignore
            maxx, maxy, maxz = v.xyz  # type: ignore
        else:
            x, y, z = v.xyz
            if x < minx:
                minx = x
            elif x > maxx:
                maxx = x
            if y < miny:
                miny = y
            elif y > maxy:
                maxy = y
            if z < minz:
                minz = z
            elif z > maxz:
                maxz = z
    if minx is None:
        raise ValueError("No vertices give.")
    return Vec3(minx, miny, minz), Vec3(maxx, maxy, maxz)


def extends2d(vertices: Iterable["Vertex"]) -> Tuple[Vec2, Vec2]:
    minx, miny = None, None
    maxx, maxy = None, None
    for v in vertices:
        v = Vec2(v)
        x, y = v.x, v.y  # type: ignore
        if minx is None:
            minx = x
            maxx = x
            miny = y
            maxy = y
        else:
            if x < minx:
                minx = x
            elif x > maxx:
                maxx = x
            if y < miny:
                miny = y
            elif y > maxy:
                maxy = y
    if minx is None:
        raise ValueError("No vertices give.")
    return Vec2(minx, miny), Vec2(maxx, maxy)



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import Iterable, Tuple
import struct

from ezdxf.tools.binarydata import BitStream
from ezdxf.entities import DXFClass

from .const import *
from .crc import crc8
from .fileheader import FileHeader
from .header_section import DwgSectionLoader


def load_classes_section(specs: FileHeader, data: Bytes, crc_check=False):
    if specs.version <= ACAD_2000:
        return DwgClassesSectionR2000(specs, data, crc_check)
    else:
        return DwgClassesSectionR2004(specs, data, crc_check)


class DwgClassesSectionR2000(DwgSectionLoader):
    def load_data_section(self, data: Bytes) -> Bytes:
        if self.specs.version > ACAD_2000:
            raise DwgVersionError(self.specs.version)
        seeker, section_size = self.specs.sections[CLASSES_ID]
        return data[seeker : seeker + section_size]

    def load_classes(self) -> Iterable[Tuple[int, DXFClass]]:
        sentinel = self.data[:SENTINEL_SIZE]
        if (
            sentinel
            != b"\x8D\xA1\xC4\xB8\xC4\xA9\xF8\xC5\xC0\xDC\xF4\x5F\xE7\xCF\xB6\x8A"
        ):
            raise DwgCorruptedClassesSection(
                "Sentinel for start of CLASSES section not found."
            )
        start_index = SENTINEL_SIZE
        bs = BitStream(
            self.data[start_index:],
            dxfversion=self.specs.version,
            encoding=self.specs.encoding,
        )
        class_data_size = bs.read_unsigned_long()  # data size in bytes
        end_sentinel_index = SENTINEL_SIZE + 6 + class_data_size
        end_index = end_sentinel_index - 2
        end_bit_index = (3 + class_data_size) << 3

        while bs.bit_index < end_bit_index:
            class_num = bs.read_bit_short()
            dxfattribs = {
                "flags": bs.read_bit_short(),  # version?
                "app_name": bs.read_text(),
                "cpp_class_name": bs.read_text(),
                "name": bs.read_text(),
                "was_a_proxy": bs.read_bit(),
                "is_an_entity": int(bs.read_bit_short() == 0x1F2),
            }
            yield class_num, DXFClass.new(dxfattribs=dxfattribs)

        if self.crc_check and False:
            check = struct.unpack_from("<H", self.data, end_index)[0]
            # TODO: classes crc check
            # Which data should be checked? This is not correct:
            crc = crc8(self.data[start_index:end_index])
            if check != crc:
                raise CRCError("CRC error in classes section.")
        sentinel = self.data[
            end_sentinel_index : end_sentinel_index + SENTINEL_SIZE
        ]
        if (
            sentinel
            != b"\x72\x5E\x3B\x47\x3B\x56\x07\x3A\x3F\x23\x0B\xA0\x18\x30\x49\x75"
        ):
            raise DwgCorruptedClassesSection(
                "Sentinel for end of CLASSES section not found."
            )


class DwgClassesSectionR2004(DwgClassesSectionR2000):
    def load_data(self, data: Bytes) -> Bytes:
        raise NotImplementedError()



# Copyright (c) 2010-2021, Manfred Moitzi
# License: MIT License
"""MText - MultiLine-Text-Entity, build by simple TEXT-Entities.

MTEXT was introduced in R13, so this is a replacement with multiple simple
TEXT entities. Supports valign (TOP, MIDDLE, BOTTOM), halign (LEFT, CENTER,
RIGHT), rotation for an arbitrary (!) angle and mirror.

This add-on exist only for porting 'dxfwrite' projects to 'ezdxf'.

"""
from typing import TYPE_CHECKING
import math
from .mixins import SubscriptAttributes
import ezdxf
from ezdxf.lldxf import const

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, GenericLayoutType


class MText(SubscriptAttributes):
    """MultiLine-Text buildup with simple Text-Entities.

    Caution: align point is always the insert point, I don't need a second
    alignpoint because horizontal alignment FIT, ALIGN, BASELINE_MIDDLE is not
    supported.

    linespacing -- line spacing in percent of height, 1.5 = 150% = 1+1/2 lines

    supported align values:
        'BOTTOM_LEFT', 'BOTTOM_CENTER', 'BOTTOM_RIGHT'
        'MIDDLE_LEFT', 'MIDDLE_CENTER', 'MIDDLE_RIGHT'
        'TOP_LEFT',    'TOP_CENTER',    'TOP_RIGHT'

    """

    MIRROR_X = const.MIRROR_X
    MIRROR_Y = const.MIRROR_Y
    TOP = const.TOP
    MIDDLE = const.MIDDLE
    BOTTOM = const.BOTTOM
    LEFT = const.LEFT
    CENTER = const.CENTER
    RIGHT = const.RIGHT
    VALID_ALIGN = frozenset(
        [
            "BOTTOM_LEFT",
            "BOTTOM_CENTER",
            "BOTTOM_RIGHT",
            "MIDDLE_LEFT",
            "MIDDLE_CENTER",
            "MIDDLE_RIGHT",
            "TOP_LEFT",
            "TOP_CENTER",
            "TOP_RIGHT",
        ]
    )

    def __init__(
        self, text: str, insert: "Vertex", linespacing: float = 1.5, **kwargs
    ):
        self.textlines = text.split("\n")
        self.insert = insert
        self.linespacing = linespacing
        if "align" in kwargs:
            self.align = kwargs.get("align", "TOP_LEFT").upper()
        else:  # support for compatibility: valign, halign
            halign = kwargs.get("halign", 0)
            valign = kwargs.get("valign", 3)
            self.align = const.TEXT_ALIGNMENT_BY_FLAGS.get(
                (halign, valign), "TOP_LEFT"
            )

        if self.align not in MText.VALID_ALIGN:
            raise ezdxf.DXFValueError(f"Invalid align parameter: {self.align}")

        self.height = kwargs.get("height", 1.0)
        self.style = kwargs.get("style", "STANDARD")
        self.oblique = kwargs.get("oblique", 0.0)  # in degree
        self.rotation = kwargs.get("rotation", 0.0)  # in degree
        self.xscale = kwargs.get("xscale", 1.0)
        self.mirror = kwargs.get(
            "mirror", 0
        )  # renamed to text_generation_flag in ezdxf
        self.layer = kwargs.get("layer", "0")
        self.color = kwargs.get("color", const.BYLAYER)

    @property
    def lineheight(self) -> float:
        """Absolute line spacing in drawing units."""
        return self.height * self.linespacing

    def render(self, layout: "GenericLayoutType") -> None:
        """Create the DXF-TEXT entities."""
        textlines = self.textlines
        if len(textlines) > 1:
            if self.mirror & const.MIRROR_Y:
                textlines.reverse()
            for linenum, text in enumerate(textlines):
                alignpoint = self._get_align_point(linenum)
                layout.add_text(
                    text,
                    dxfattribs=self._dxfattribs(alignpoint),
                )
        elif len(textlines) == 1:
            layout.add_text(
                textlines[0],
                dxfattribs=self._dxfattribs(self.insert),
            )

    def _get_align_point(self, linenum: int) -> "Vertex":
        """Calculate the align point depending on the line number."""
        x = self.insert[0]
        y = self.insert[1]
        try:
            z = self.insert[2]
        except IndexError:
            z = 0.0
        # rotation not respected

        if self.align.startswith("TOP"):
            y -= linenum * self.lineheight
        elif self.align.startswith("MIDDLE"):
            y0 = linenum * self.lineheight
            fullheight = (len(self.textlines) - 1) * self.lineheight
            y += (fullheight / 2) - y0
        else:  # BOTTOM
            y += (len(self.textlines) - 1 - linenum) * self.lineheight
        return self._rotate((x, y, z))  # consider rotation

    def _rotate(self, alignpoint: "Vertex") -> "Vertex":
        """Rotate alignpoint around insert point about rotation degrees."""
        dx = alignpoint[0] - self.insert[0]
        dy = alignpoint[1] - self.insert[1]
        beta = math.radians(self.rotation)
        x = self.insert[0] + dx * math.cos(beta) - dy * math.sin(beta)
        y = self.insert[1] + dy * math.cos(beta) + dx * math.sin(beta)
        return round(x, 6), round(y, 6), alignpoint[2]

    def _dxfattribs(self, alignpoint: "Vertex") -> dict:
        """Build keyword arguments for TEXT entity creation."""
        halign, valign = const.TEXT_ALIGN_FLAGS.get(self.align, (0, 3))
        return {
            "insert": alignpoint,
            "align_point": alignpoint,
            "layer": self.layer,
            "color": self.color,
            "style": self.style,
            "height": self.height,
            "width": self.xscale,
            "text_generation_flag": self.mirror,
            "rotation": self.rotation,
            "oblique": self.oblique,
            "halign": halign,
            "valign": valign,
        }



# Copyright (c) 2020-2021, Matthew Broadway
# License: MIT License
import math
from abc import ABCMeta
from typing import Optional, Iterable, Dict, Union, Tuple, no_type_check
from collections import defaultdict
from functools import lru_cache
from PyQt5 import QtCore as qc, QtGui as qg, QtWidgets as qw

from ezdxf.addons.drawing.backend import Backend, prepare_string_for_rendering
from ezdxf.addons.drawing.config import Configuration, LinePolicy, HatchPolicy
from ezdxf.tools.fonts import FontMeasurements
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.addons.drawing.properties import Properties
from ezdxf.addons.drawing.line_renderer import AbstractLineRenderer
from ezdxf.tools import fonts
from ezdxf.math import Vec3, Matrix44
from ezdxf.path import Path, Command
from ezdxf.render.linetypes import LineTypeRenderer as EzdxfLineTypeRenderer
from ezdxf.tools.pattern import PatternAnalyser

PatternKey = Tuple[str, float]


class _Point(qw.QAbstractGraphicsShapeItem):
    """A dimensionless point which is drawn 'cosmetically' (scale depends on
    view)
    """

    def __init__(self, x: float, y: float, brush: qg.QBrush):
        super().__init__()
        self.location = qc.QPointF(x, y)
        self.radius = 1.0
        self.setPen(qg.QPen(qc.Qt.NoPen))
        self.setBrush(brush)

    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        view_scale = _get_x_scale(painter.transform())
        radius = self.radius / view_scale
        painter.setBrush(self.brush())
        painter.setPen(qc.Qt.NoPen)
        painter.drawEllipse(self.location, radius, radius)

    def boundingRect(self) -> qc.QRectF:
        return qc.QRectF(self.location, qc.QSizeF(1, 1))


# The key used to store the dxf entity corresponding to each graphics element
CorrespondingDXFEntity = qc.Qt.UserRole + 0
CorrespondingDXFParentStack = qc.Qt.UserRole + 1


class PyQtBackend(Backend):
    def __init__(
        self,
        scene: Optional[qw.QGraphicsScene] = None,
        *,
        use_text_cache: bool = True,
        debug_draw_rect: bool = False,
        extra_lineweight_scaling: float = 2.0,
    ):
        """
        Args:
            extra_lineweight_scaling: compared to other backends,
                PyQt draws lines which appear thinner
        """
        super().__init__()
        self._scene = scene or qw.QGraphicsScene()  # avoids many type errors
        self._color_cache: Dict[Color, qg.QColor] = {}
        self._pattern_cache: Dict[PatternKey, int] = {}
        self._no_line = qg.QPen(qc.Qt.NoPen)
        self._no_fill = qg.QBrush(qc.Qt.NoBrush)

        self._text_renderer = TextRenderer(qg.QFont(), use_text_cache)
        self._line_renderer: PyQtLineRenderer
        self._extra_lineweight_scaling = extra_lineweight_scaling
        self._debug_draw_rect = debug_draw_rect

    def configure(self, config: Configuration) -> None:
        if config.min_lineweight is None:
            config = config.with_changes(min_lineweight=0.24)
        super().configure(config)
        if config.line_policy == LinePolicy.SOLID:
            self._line_renderer = InternalLineRenderer(
                self, config, solid_only=True
            )
        elif config.line_policy == LinePolicy.APPROXIMATE:
            self._line_renderer = InternalLineRenderer(
                self, config, solid_only=False
            )
        elif config.line_policy == LinePolicy.ACCURATE:
            self._line_renderer = EzdxfLineRenderer(self, config)
        else:
            raise ValueError(config.line_policy)

    def set_scene(self, scene: qw.QGraphicsScene):
        self._scene = scene

    def clear_text_cache(self):
        self._text_renderer.clear_cache()

    def _get_color(self, color: Color) -> qg.QColor:
        qt_color = self._color_cache.get(color, None)
        if qt_color is None:
            if len(color) == 7:
                qt_color = qg.QColor(color)  # '#RRGGBB'
            elif len(color) == 9:
                rgb = color[1:7]
                alpha = color[7:9]
                qt_color = qg.QColor(f"#{alpha}{rgb}")  # '#AARRGGBB'
            else:
                raise TypeError(color)

            self._color_cache[color] = qt_color
        return qt_color

    def _get_pen(self, properties: Properties) -> qg.QPen:
        """Returns a cosmetic pen with applied lineweight but without line type
        support.
        """
        px = (
            properties.lineweight
            / 0.3527
            * self.config.lineweight_scaling
            * self._extra_lineweight_scaling
        )
        pen = qg.QPen(self._get_color(properties.color), px)
        # Use constant width in pixel:
        pen.setCosmetic(True)
        pen.setJoinStyle(qc.Qt.RoundJoin)
        return pen

    def _get_brush(self, properties: Properties) -> qg.QBrush:
        filling = properties.filling
        if filling:
            if filling.type == filling.PATTERN:
                if (
                    self.config.hatch_policy
                    == HatchPolicy.SHOW_APPROXIMATE_PATTERN
                ):
                    # Default pattern scaling is not supported by PyQt:
                    key: PatternKey = (filling.name, filling.angle)
                    qt_pattern = self._pattern_cache.get(key)  # type: ignore
                    if qt_pattern is None:
                        qt_pattern = self._get_qt_pattern(filling.pattern)
                        self._pattern_cache[key] = qt_pattern
                elif self.config.hatch_policy == HatchPolicy.SHOW_SOLID:
                    qt_pattern = qc.Qt.SolidPattern
                elif self.config.hatch_policy == HatchPolicy.SHOW_OUTLINE:
                    return self._no_fill
                else:
                    raise ValueError(self.config.hatch_policy)
            else:
                qt_pattern = qc.Qt.SolidPattern

            return qg.QBrush(self._get_color(properties.color), qt_pattern)  # type: ignore
        else:
            return self._no_fill

    @staticmethod
    def _get_qt_pattern(pattern) -> int:
        pattern = PatternAnalyser(pattern)
        # knowledge of dark or light background would by handy:
        qt_pattern = qc.Qt.Dense4Pattern
        if pattern.all_angles(0):
            qt_pattern = qc.Qt.HorPattern
        elif pattern.all_angles(90):
            qt_pattern = qc.Qt.VerPattern
        elif pattern.has_angle(0) and pattern.has_angle(90):
            qt_pattern = qc.Qt.CrossPattern
        if pattern.all_angles(45):
            qt_pattern = qc.Qt.BDiagPattern
        elif pattern.all_angles(135):
            qt_pattern = qc.Qt.FDiagPattern
        elif pattern.has_angle(45) and pattern.has_angle(135):
            qt_pattern = qc.Qt.DiagCrossPattern
        return qt_pattern

    def _set_item_data(self, item: qw.QGraphicsItem) -> None:
        parent_stack = tuple(e for e, props in self.entity_stack[:-1])
        current_entity = self.current_entity
        if isinstance(item, list):
            for item_ in item:
                item_.setData(CorrespondingDXFEntity, current_entity)
                item_.setData(CorrespondingDXFParentStack, parent_stack)
        else:
            item.setData(CorrespondingDXFEntity, current_entity)
            item.setData(CorrespondingDXFParentStack, parent_stack)

    def set_background(self, color: Color):
        self._scene.setBackgroundBrush(qg.QBrush(self._get_color(color)))

    def draw_point(self, pos: Vec3, properties: Properties) -> None:
        """Draw a real dimensionless point."""
        brush = qg.QBrush(self._get_color(properties.color), qc.Qt.SolidPattern)
        item = _Point(pos.x, pos.y, brush)
        self._set_item_data(item)
        self._scene.addItem(item)

    def draw_line(self, start: Vec3, end: Vec3, properties: Properties) -> None:
        # PyQt draws a long line for a zero-length line:
        if start.isclose(end):
            self.draw_point(start, properties)
        else:
            item = self._line_renderer.draw_line(start, end, properties)  # type: ignore
            self._set_item_data(item)

    def draw_path(self, path: Path, properties: Properties) -> None:
        item = self._line_renderer.draw_path(path, properties)  # type: ignore
        self._set_item_data(item)

    def draw_filled_paths(
        self,
        paths: Iterable[Path],
        holes: Iterable[Path],
        properties: Properties,
    ) -> None:
        qt_path = qg.QPainterPath()
        for path in paths:
            try:
                path = path.counter_clockwise()
            except ValueError:  # cannot detect path orientation
                continue
            _extend_qt_path(qt_path, path)
        for path in holes:
            try:
                path = path.clockwise()
            except ValueError:  # cannot detect path orientation
                continue
            _extend_qt_path(qt_path, path)
        item = _CosmeticPath(qt_path)
        item.setPen(self._get_pen(properties))
        item.setBrush(self._get_brush(properties))
        self._scene.addItem(item)
        self._set_item_data(item)

    def draw_filled_polygon(
        self, points: Iterable[Vec3], properties: Properties
    ) -> None:
        brush = self._get_brush(properties)
        polygon = qg.QPolygonF()
        for p in points:
            polygon.append(qc.QPointF(p.x, p.y))
        item = _CosmeticPolygon(polygon)
        item.setPen(self._no_line)
        item.setBrush(brush)
        self._scene.addItem(item)
        self._set_item_data(item)

    def draw_text(
        self,
        text: str,
        transform: Matrix44,
        properties: Properties,
        cap_height: float,
    ) -> None:
        if not text.strip():
            return  # no point rendering empty strings
        text = prepare_string_for_rendering(text, self.current_entity.dxftype())  # type: ignore
        qfont = self.get_qfont(properties.font)
        scale = self._text_renderer.get_scale(cap_height, qfont)
        transform = Matrix44.scale(scale, -scale, 0) @ transform

        path = self._text_renderer.get_text_path(text, qfont)
        path = _matrix_to_qtransform(transform).map(path)
        item = self._scene.addPath(
            path, self._no_line, self._get_color(properties.color)
        )
        self._set_item_data(item)

    @lru_cache(maxsize=256)  # fonts.Font is a named tuple
    def get_qfont(self, font: fonts.FontFace) -> qg.QFont:
        qfont = self._text_renderer.default_font
        if font:
            family = font.family
            italic = "italic" in font.style.lower()
            weight = _map_weight(font.weight)
            qfont = qg.QFont(family, weight=weight, italic=italic)
            # INFO: setting the stretch value makes results worse!
            # qfont.setStretch(_map_stretch(font.stretch))
        return qfont

    def get_font_measurements(
        self, cap_height: float, font: fonts.FontFace = None
    ) -> FontMeasurements:
        qfont = self.get_qfont(font)
        return self._text_renderer.get_font_measurements(
            qfont
        ).scale_from_baseline(desired_cap_height=cap_height)

    def get_text_line_width(
        self, text: str, cap_height: float, font: fonts.FontFace = None
    ) -> float:
        if not text.strip():
            return 0

        dxftype = (
            self.current_entity.dxftype() if self.current_entity else "TEXT"
        )
        text = prepare_string_for_rendering(text, dxftype)
        qfont = self.get_qfont(font)
        scale = self._text_renderer.get_scale(cap_height, qfont)
        return self._text_renderer.get_text_rect(text, qfont).right() * scale

    def clear(self) -> None:
        self._scene.clear()

    def finalize(self) -> None:
        super().finalize()
        self._scene.setSceneRect(self._scene.itemsBoundingRect())
        if self._debug_draw_rect:
            properties = Properties()
            properties.color = "#000000"
            self._scene.addRect(
                self._scene.sceneRect(),
                self._get_pen(properties),
                self._no_fill,
            )


class _CosmeticPath(qw.QGraphicsPathItem):
    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        _set_cosmetic_brush(self, painter)
        super().paint(painter, option, widget)


class _CosmeticPolygon(qw.QGraphicsPolygonItem):
    def paint(
        self,
        painter: qg.QPainter,
        option: qw.QStyleOptionGraphicsItem,
        widget: Optional[qw.QWidget] = None,
    ) -> None:
        _set_cosmetic_brush(self, painter)
        super().paint(painter, option, widget)


def _set_cosmetic_brush(
    item: qw.QAbstractGraphicsShapeItem, painter: qg.QPainter
) -> None:
    """like a cosmetic pen, this sets the brush pattern to appear the same independent of the view"""
    brush = item.brush()
    # scale by -1 in y because the view is always mirrored in y and undoing the view transformation entirely would make
    # the hatch mirrored w.r.t the view
    brush.setTransform(painter.transform().inverted()[0].scale(1, -1))
    item.setBrush(brush)


@no_type_check
def _extend_qt_path(qt_path: qg.QPainterPath, path: Path) -> None:
    start = path.start
    qt_path.moveTo(start.x, start.y)
    for cmd in path:
        if cmd.type == Command.LINE_TO:
            end = cmd.end
            qt_path.lineTo(end.x, end.y)
        elif cmd.type == Command.CURVE4_TO:
            end = cmd.end
            ctrl1 = cmd.ctrl1
            ctrl2 = cmd.ctrl2
            qt_path.cubicTo(ctrl1.x, ctrl1.y, ctrl2.x, ctrl2.y, end.x, end.y)
        else:
            raise ValueError(f"Unknown path command: {cmd.type}")


# https://doc.qt.io/qt-5/qfont.html#Weight-enum
# QFont::Thin	0	0
# QFont::ExtraLight	12	12
# QFont::Light	25	25
# QFont::Normal	50	50
# QFont::Medium	57	57
# QFont::DemiBold	63	63
# QFont::Bold	75	75
# QFont::ExtraBold	81	81
# QFont::Black	87	87
def _map_weight(weight: Union[str, int]) -> int:
    if isinstance(weight, str):
        weight = fonts.weight_name_to_value(weight)
    value = int((weight / 10) + 10)  # normal: 400 -> 50
    return min(max(0, value), 99)


# https://doc.qt.io/qt-5/qfont.html#Stretch-enum
StretchMapping = {
    "ultracondensed": 50,
    "extracondensed": 62,
    "condensed": 75,
    "semicondensed": 87,
    "unstretched": 100,
    "semiexpanded": 112,
    "expanded": 125,
    "extraexpanded": 150,
    "ultraexpanded": 200,
}


def _map_stretch(stretch: str) -> int:
    return StretchMapping.get(stretch.lower(), 100)


def _get_x_scale(t: qg.QTransform) -> float:
    return math.sqrt(t.m11() * t.m11() + t.m21() * t.m21())


def _matrix_to_qtransform(matrix: Matrix44) -> qg.QTransform:
    """Qt also uses row-vectors so the translation elements are placed in the
    bottom row.

    This is only a simple conversion which assumes that although the
    transformation is 4x4,it does not involve the z axis.

    A more correct transformation could be implemented like so:
    https://stackoverflow.com/questions/10629737/convert-3d-4x4-rotation-matrix-into-2d
    """
    return qg.QTransform(*matrix.get_2d_transformation())


class TextRenderer:
    def __init__(self, font: qg.QFont, use_cache: bool):
        self._default_font = font
        self._use_cache = use_cache

        # Each font has its own text path cache
        # key is QFont.key()
        self._text_path_cache: Dict[
            str, Dict[str, qg.QPainterPath]
        ] = defaultdict(dict)

        # Each font has its own font measurements cache
        # key is QFont.key()
        self._font_measurement_cache: Dict[str, FontMeasurements] = {}

    @property
    def default_font(self) -> qg.QFont:
        return self._default_font

    def clear_cache(self):
        self._text_path_cache.clear()

    def get_scale(self, desired_cap_height: float, font: qg.QFont) -> float:
        measurements = self.get_font_measurements(font)
        return desired_cap_height / measurements.cap_height

    def get_font_measurements(self, font: qg.QFont) -> FontMeasurements:
        # None is the default font.
        key = font.key() if font is not None else None
        measurements = self._font_measurement_cache.get(key)
        if measurements is None:
            upper_x = self.get_text_rect("X", font)
            lower_x = self.get_text_rect("x", font)
            lower_p = self.get_text_rect("p", font)
            baseline = lower_x.bottom()
            measurements = FontMeasurements(
                baseline=baseline,
                cap_height=baseline - upper_x.top(),
                x_height=baseline - lower_x.top(),
                descender_height=lower_p.bottom() - baseline,
            )
            self._font_measurement_cache[key] = measurements
        return measurements

    def get_text_path(self, text: str, font: qg.QFont) -> qg.QPainterPath:
        # None is the default font
        key = font.key() if font is not None else None
        cache = self._text_path_cache[key]  # defaultdict(dict)
        path = cache.get(text, None)
        if path is None:
            if font is None:
                font = self._default_font
            path = qg.QPainterPath()
            path.addText(0, 0, font, text)
            if self._use_cache:
                cache[text] = path
        return path

    def get_text_rect(self, text: str, font: qg.QFont) -> qc.QRectF:
        # no point caching the bounding rect calculation, it is very cheap
        return self.get_text_path(text, font).boundingRect()


# noinspection PyProtectedMember
class PyQtLineRenderer(AbstractLineRenderer, metaclass=ABCMeta):
    def __init__(self, backend: PyQtBackend, config: Configuration):
        super().__init__(config)
        self._backend = backend

    @property
    def scene(self) -> qw.QGraphicsScene:
        return self._backend._scene

    @property
    def no_fill(self) -> qg.QBrush:
        return self._backend._no_fill

    def get_color(self, color: Color) -> qg.QColor:
        return self._backend._get_color(color)

    def get_pen(self, properties: Properties) -> qg.QPen:
        return self._backend._get_pen(properties)


# Just guessing here: this values assume a cosmetic pen!
ISO_LIN_PATTERN_FACTOR = 15
ANSI_LIN_PATTERN_FACTOR = ISO_LIN_PATTERN_FACTOR * 2.54


class InternalLineRenderer(PyQtLineRenderer):
    """PyQt internal linetype rendering"""

    def __init__(
        self, backend: PyQtBackend, config: Configuration, solid_only: bool
    ):
        super().__init__(backend, config)
        self._solid_only = solid_only

    @property
    def measurement_scale(self) -> float:
        return (
            ISO_LIN_PATTERN_FACTOR
            if self._config.measurement
            else ISO_LIN_PATTERN_FACTOR
        )

    def get_pen(self, properties: Properties) -> qg.QPen:
        pen = super().get_pen(properties)
        if not self._solid_only and len(properties.linetype_pattern) > 1:
            # The dash pattern is specified in units of the pens width; e.g. a
            # dash of length 5 in width 10 is 50 pixels long.
            pattern = self.pattern(properties)
            if len(pattern):
                pen.setDashPattern(pattern)
        return pen

    def draw_line(self, start: Vec3, end: Vec3, properties: Properties, z=0):
        return self.scene.addLine(
            start.x, start.y, end.x, end.y, self.get_pen(properties)
        )

    def draw_path(self, path: Path, properties: Properties, z=0):
        qt_path = qg.QPainterPath()
        _extend_qt_path(qt_path, path)
        return self.scene.addPath(
            qt_path,
            self.get_pen(properties),
            self.no_fill,
        )


class EzdxfLineRenderer(PyQtLineRenderer):
    """Replicate AutoCAD linetype rendering oriented on drawing units and
    various ltscale factors. This rendering method break lines into small
    segments which causes a longer rendering time!
    """

    def draw_line(self, start: Vec3, end: Vec3, properties: Properties, z=0):
        pattern = self.pattern(properties)
        pen = self.get_pen(properties)
        if len(pattern) < 2:
            return self.scene.addLine(start.x, start.y, end.x, end.y, pen)
        else:
            add_line = self.scene.addLine
            renderer = EzdxfLineTypeRenderer(pattern)
            return [
                add_line(s.x, s.y, e.x, e.y, pen)
                for s, e in renderer.line_segment(start, end)
                # PyQt has problems with very short lines:
                if not s.isclose(e)
            ]

    def draw_path(self, path, properties: Properties, z=0):
        pattern = self.pattern(properties)
        pen = self.get_pen(properties)
        if len(pattern) < 2:
            qt_path = qg.QPainterPath()
            _extend_qt_path(qt_path, path)
            return self.scene.addPath(qt_path, pen, self.no_fill)
        else:
            add_line = self.scene.addLine
            renderer = EzdxfLineTypeRenderer(pattern)
            segments = renderer.line_segments(
                path.flattening(
                    self._config.max_flattening_distance, segments=16
                )
            )
            return [
                add_line(s.x, s.y, e.x, e.y, pen)
                for s, e in segments
                # PyQt has problems with very short lines:
                if not s.isclose(e)
            ]



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
import logging
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER
from ezdxf.lldxf import validator
from ezdxf.math import UCS, NULLVEC, X_AXIS, Y_AXIS
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from ezdxf.entities.layer import acdb_symbol_table_record
from .factory import register_entity

logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ["UCSTableEntry"]

acdb_ucs = DefSubclass(
    "AcDbUCSTableRecord",
    {
        "name": DXFAttr(2, validator=validator.is_valid_table_name),
        "flags": DXFAttr(70, default=0),
        "origin": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        "xaxis": DXFAttr(
            11,
            xtype=XType.point3d,
            default=X_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        "yaxis": DXFAttr(
            12,
            xtype=XType.point3d,
            default=Y_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_ucs_group_codes = group_code_mapping(acdb_ucs)


@register_entity
class UCSTableEntry(DXFEntity):
    """DXF UCS table entity"""

    DXFTYPE = "UCS"
    DXFATTRIBS = DXFAttributes(base_class, acdb_symbol_table_record, acdb_ucs)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_ucs_group_codes, subclass=2
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_ucs.name)

        self.dxf.export_dxf_attribs(
            tagwriter, ["name", "flags", "origin", "xaxis", "yaxis"]
        )

    def ucs(self) -> UCS:
        """Returns an :class:`ezdxf.math.UCS` object for this UCS table entry."""
        return UCS(
            origin=self.dxf.origin,
            ux=self.dxf.xaxis,
            uy=self.dxf.yaxis,
        )



# Purpose: DXF Pretty Printer
# Copyright (c) 2015-2021, Manfred Moitzi
# License: MIT License
from typing import Iterable
import sys
import io
from pathlib import Path

from .dxfpp import dxfpp
from .rawpp import rawpp
from ezdxf.lldxf.const import DXFError, DXFStructureError
from ezdxf.lldxf.tagger import (
    ascii_tags_loader,
    tag_compiler,
    binary_tags_loader,
)
from ezdxf.lldxf.types import DXFTag
from ezdxf.lldxf.validator import is_dxf_file, is_binary_dxf_file
from ezdxf.filemanagement import dxf_file_info
from ezdxf.lldxf.repair import tag_reorder_layer
import webbrowser


def readfile(
    filename: str,
    legacy_mode: bool = False,
    compile_tags=True,
    is_binary_dxf=False,
) -> Iterable[DXFTag]:
    if is_binary_dxf:
        with open(filename, mode="rb") as fp:
            data = fp.read()
            tagger = binary_tags_loader(data)
    else:
        info = dxf_file_info(filename)
        fp = open(filename, mode="rt", encoding=info.encoding, errors="ignore")  # type: ignore
        tagger = ascii_tags_loader(fp)  # type: ignore

    if legacy_mode:
        tagger = tag_reorder_layer(tagger)
    if compile_tags:
        tagger = tag_compiler(iter(tagger))
    return tagger


def pretty_print(filename: Path, is_binary_dxf=False):
    try:
        tagger = readfile(
            str(filename), legacy_mode=True, is_binary_dxf=is_binary_dxf
        )
    except IOError:
        print(f'Unable to read DXF file "{filename}".')
        sys.exit(1)
    except DXFError as e:
        print(str(e))
        sys.exit(2)

    html_filename = filename.parent / (filename.stem + ".html")
    try:
        with io.open(html_filename, mode="wt", encoding="utf-8") as fp:
            fp.write(dxfpp(tagger, filename.name))
    except IOError:
        print(f'IOError: can not write file "{html_filename}".')
    return html_filename


def raw_pretty_print(
    filename: Path, compile_tags=True, legacy_mode=False, is_binary_dxf=False
):
    try:
        tagger = readfile(
            str(filename),
            legacy_mode=legacy_mode,
            compile_tags=compile_tags,
            is_binary_dxf=is_binary_dxf,
        )
    except IOError:
        print(f'Unable to read DXF file "{filename}".')
        sys.exit(1)
    except DXFError as e:
        print(str(e))
        sys.exit(2)

    html_filename = filename.parent / (filename.stem + ".html")
    try:
        with io.open(html_filename, mode="wt", encoding="utf-8") as html:
            html.write(rawpp(tagger, str(filename), binary=is_binary_dxf))
    except IOError:
        print(f'IOError: can not write file "{filename}".')
    except DXFStructureError as e:
        print(f"DXFStructureError: {str(e)}")
    return html_filename


def run(args):
    for filename in args.files:
        if not Path(filename).exists():
            print(f'File "{filename}" not found.')
            continue

        if is_binary_dxf_file(filename):
            binary_dxf = True
        else:
            binary_dxf = False
            if not is_dxf_file(filename):
                print(f'File "{filename}" is not a DXF file.')
                continue

        if args.raw:
            html_path = raw_pretty_print(
                Path(filename),
                compile_tags=not args.nocompile,
                legacy_mode=args.legacy,
                is_binary_dxf=binary_dxf,
            )
        else:
            html_path = pretty_print(
                Path(filename), is_binary_dxf=binary_dxf
            )  # legacy mode is always used

        print(f'Created "{html_path}"')
        if args.open:
            webbrowser.open(str(html_path))
